"""
This module provides functionality for calculating the Common Language Effect Size (CLES) for one sample t-tests.

Classes:
    OneSampleCLESResults: A class containing results from CLES calculations
    OneSampleTTest: A class containing static methods for calculating the Common Language Effect Size (CLES) and other related statistics.

Methods:
    pivotal_ci_t: Calculate the pivotal confidence intervals for a t-score.
    calculate_central_ci_from_cohens_d_one_sample_t_test: Calculate the central confidence intervals for Cohen's d in a one-sample t-test.
    CI_NCP_one_Sample: Calculate the Non-Central Parameter (NCP) confidence intervals for a one-sample t-test.
    density: Calculate the density function for a given value.
    area_under_function: Calculate the area under a given function using numerical integration.
    WinsorizedVariance: Calculate the Winsorized variance of a sample.
    WinsorizedCorrelation: Calculate the Winsorized correlation between two samples.
"""

import math
from typing import Optional
from dataclasses import dataclass
import numpy as np
import numpy.typing as npt
from scipy.stats import norm, nct, t, trim_mean


@dataclass
class OneSampleCLESResults:
    """Class for storing one-sample CLES test results"""

    # Common Language Effect Sizes
    cles_d: Optional[npt.NDArray[np.float16]] = (
        None  # McGraw & Wong Common Language Effect Size
    )
    cles_g: Optional[npt.NDArray[np.float16]] = (
        None  # McGraw & Wong Unbiased Common Language Effect Size
    )

    # Test Statistics
    t_score: Optional[npt.NDArray[np.float16]] = None
    df: Optional[npt.NDArray[np.float16]] = None
    p_value: Optional[npt.NDArray[np.float16]] = None

    # Central CI's
    lower_central_ci_cld: Optional[npt.NDArray[np.float16]] = None
    upper_central_ci_cld: Optional[npt.NDArray[np.float16]] = None
    lower_central_ci_clg: Optional[npt.NDArray[np.float16]] = None
    upper_central_ci_clg: Optional[npt.NDArray[np.float16]] = None

    # Non-Central CI's
    lower_ncp_ci_cld: Optional[npt.NDArray[np.float16]] = None
    upper_ncp_ci_cld: Optional[npt.NDArray[np.float16]] = None
    lower_ncp_ci_clg: Optional[npt.NDArray[np.float16]] = None
    upper_ncp_ci_clg: Optional[npt.NDArray[np.float16]] = None

    # Pivotal CI's
    lower_pivotal_ci_cld: Optional[npt.NDArray[np.float16]] = None
    upper_pivotal_ci_cld: Optional[npt.NDArray[np.float16]] = None
    lower_pivotal_ci_clg: Optional[npt.NDArray[np.float16]] = None
    upper_pivotal_ci_clg: Optional[npt.NDArray[np.float16]] = None

    # Robust Statistics
    robust_effect_size_akp: Optional[npt.NDArray[np.float16]] = None
    lower_ci_robust_akp: Optional[npt.NDArray[np.float16]] = None
    upper_ci_robust_akp: Optional[npt.NDArray[np.float16]] = None
    trimmed_mean: Optional[npt.NDArray[np.float16] | float] = None
    winsorized_sd: Optional[npt.NDArray[np.float16]] = None
    yuens_t: Optional[npt.NDArray[np.float16]] = None
    robust_df: Optional[npt.NDArray[np.float16]] = None
    robust_p_value: Optional[npt.NDArray[np.float16]] = None
    mean_difference: Optional[npt.NDArray[np.float16]] = None
    standard_error: Optional[npt.NDArray[np.float16]] = None
    statistical_line_robust_akp: Optional[str] = None
    statistical_line_cld: Optional[str] = None
    statistical_line_clg: Optional[str] = None


def pivotal_ci_t(t_score, df, sample_size, confidence_level):
    """
    Calculate the pivotal confidence intervals for a t-score.

    Parameters
    ----------
    t_Score : float
        The t-score value.
    df : int
        Degrees of freedom.
    sample_size : int
        The size of the sample.
    confidence_level : float
        The confidence level as a decimal (e.g., 0.95 for 95%).

    Returns
    -------
    tuple
        A tuple containing:
        - lower_ci (float): Lower bound of the pivotal confidence interval.
        - upper_ci (float): Upper bound of the pivotal confidence interval.
    """
    is_negative = False
    if t_score < 0:
        is_negative = True
        t_score = abs(t_score)
    upper_limit = 1 - (1 - confidence_level) / 2
    lower_limit = (1 - confidence_level) / 2

    lower_criterion = [-t_score, t_score / 2, t_score]
    upper_criterion = [t_score, 2 * t_score, 3 * t_score]

    while nct.cdf(t_score, df, lower_criterion[0]) < upper_limit:
        lower_criterion = [
            lower_criterion[0] - t_score,
            lower_criterion[0],
            lower_criterion[2],
        ]

    while nct.cdf(t_score, df, upper_criterion[0]) < lower_limit:
        if nct.cdf(t_score, df) < lower_limit:
            lower_ci = [0, nct.cdf(t_score, df)]
            upper_criterion = [
                upper_criterion[0] / 4,
                upper_criterion[0],
                upper_criterion[2],
            ]

    while nct.cdf(t_score, df, upper_criterion[2]) > lower_limit:
        upper_criterion = [
            upper_criterion[0],
            upper_criterion[2],
            upper_criterion[2] + t_score,
        ]

    lower_ci = 0.0
    diff_lower = 1
    while diff_lower > 0.00001:
        if nct.cdf(t_score, df, lower_criterion[1]) < upper_limit:
            lower_criterion = [
                lower_criterion[0],
                (lower_criterion[0] + lower_criterion[1]) / 2,
                lower_criterion[1],
            ]
        else:
            lower_criterion = [
                lower_criterion[1],
                (lower_criterion[1] + lower_criterion[2]) / 2,
                lower_criterion[2],
            ]
        diff_lower = abs(nct.cdf(t_score, df, lower_criterion[1]) - upper_limit)
        lower_ci = lower_criterion[1] / (np.sqrt(sample_size))

    upper_ci = 0.0
    diff_upper = 1
    while diff_upper > 0.00001:
        if nct.cdf(t_score, df, upper_criterion[1]) < lower_limit:
            upper_criterion = [
                upper_criterion[0],
                (upper_criterion[0] + upper_criterion[1]) / 2,
                upper_criterion[1],
            ]
        else:
            upper_criterion = [
                upper_criterion[1],
                (upper_criterion[1] + upper_criterion[2]) / 2,
                upper_criterion[2],
            ]
        diff_upper = abs(nct.cdf(t_score, df, upper_criterion[1]) - lower_limit)
        upper_ci = upper_criterion[1] / (np.sqrt(sample_size))
    if is_negative:
        return -upper_ci, -lower_ci
    else:
        return lower_ci, upper_ci


def calculate_central_ci_from_cohens_d_one_sample_t_test(
    cohens_d, sample_size, confidence_level
):
    """
    Calculate the central confidence intervals for Cohen's d in a one-sample t-test.

    Parameters
    ----------
    cohens_d : float
        The calculated Cohen's d effect size.
    sample_size : int
        The size of the sample.
    confidence_level : float
        The confidence level as a decimal (e.g., 0.95 for 95%).

    Returns
    -------
    tuple
        A tuple containing:
        - ci_lower (float): Lower bound of the central confidence interval.
        - ci_upper (float): Upper bound of the central confidence interval.
        - standard_error_es (float): Standard error of the effect size.
    """
    df = sample_size - 1
    correction_factor = math.exp(
        math.lgamma(df / 2) - math.log(math.sqrt(df / 2)) - math.lgamma((df - 1) / 2)
    )
    standard_error_es = np.sqrt(
        (df / (df - 2)) * (1 / sample_size) * (1 + cohens_d**2 * sample_size)
        - (cohens_d**2 / correction_factor**2)
    )
    z_critical_value = norm.ppf(confidence_level + ((1 - confidence_level) / 2))
    ci_lower, ci_upper = (
        cohens_d - standard_error_es * z_critical_value,
        cohens_d + standard_error_es * z_critical_value,
    )
    return ci_lower, ci_upper, standard_error_es


def ci_ncp_one_sample(effect_size, sample_size, confidence_level):
    """
    Calculate the Non-Central Parameter (NCP) confidence intervals for a one-sample t-test.

    Parameters
    ----------
    Effect_Size : float
        The calculated effect size (Cohen's d).
    sample_size : int
        The size of the sample.
    confidence_level : float
        The confidence level as a decimal (e.g., 0.95 for 95%).

    Returns
    -------
    tuple
        A tuple containing:
        - CI_NCP_low (float): Lower bound of the NCP confidence interval.
        - CI_NCP_High (float): Upper bound of the NCP confidence interval.
    """
    NCP_value = effect_size * math.sqrt(sample_size)
    CI_NCP_low = (
        (
            nct.ppf(
                1 / 2 - confidence_level / 2,
                (sample_size - 1),
                loc=0,
                scale=1,
                nc=NCP_value,
            )
        )
        / NCP_value
        * effect_size
    )
    CI_NCP_High = (
        (
            nct.ppf(
                1 / 2 + confidence_level / 2,
                (sample_size - 1),
                loc=0,
                scale=1,
                nc=NCP_value,
            )
        )
        / NCP_value
        * effect_size
    )
    return CI_NCP_low, CI_NCP_High


def density(x):
    """
    Calculate the density function for a given value.

    Parameters
    ----------
    x : float
        The input value.

    Returns
    -------
    float
        The density value.
    """
    x = np.array(x)
    return x**2 * norm.pdf(x)


def area_under_function(
    f, a, b, *args, function_a=None, function_b=None, limit=10, eps=1e-5
):
    """
    Calculate the area under a given function using numerical integration.

    Parameters
    ----------
    f : function
        The function to integrate.
    a : float
        The lower bound of the integration.
    b : float
        The upper bound of the integration.
    *args : tuple
        Additional arguments to pass to the function.
    function_a : float, optional
        The function value at the lower bound.
    function_b : float, optional
        The function value at the upper bound.
    limit : int, optional
        The maximum number of recursive calls (default is 10).
    eps : float, optional
        The tolerance for stopping the recursion (default is 1e-5).

    Returns
    -------
    float
        The area under the function.
    """
    if function_a is None:
        function_a = f(a, *args)
    if function_b is None:
        function_b = f(b, *args)
    midpoint = (a + b) / 2
    f_midpoint = f(midpoint, *args)
    area_trapezoidal = ((function_a + function_b) * (b - a)) / 2
    area_simpson = ((function_a + 4 * f_midpoint + function_b) * (b - a)) / 6
    if abs(area_trapezoidal - area_simpson) < eps or limit == 0:
        return area_simpson
    return area_under_function(
        f,
        a,
        midpoint,
        *args,
        function_a=function_a,
        function_b=f_midpoint,
        limit=limit - 1,
        eps=eps,
    ) + area_under_function(
        f,
        midpoint,
        b,
        *args,
        function_a=f_midpoint,
        function_b=function_b,
        limit=limit - 1,
        eps=eps,
    )


def WinsorizedVariance(x, trimming_level=0.2):
    """
    Calculate the Winsorized variance of a sample.

    Parameters
    ----------
    x : array-like
        The input sample.
    trimming_level : float, optional
        The trimming level (default is 0.2).

    Returns
    -------
    float
        The Winsorized variance.
    """
    y = np.sort(x)
    n = len(x)
    ibot = int(np.floor(trimming_level * n)) + 1
    itop = n - ibot + 1
    xbot = y[ibot - 1]
    xtop = y[itop - 1]
    y = np.where(y <= xbot, xbot, y)
    y = np.where(y >= xtop, xtop, y)
    winvar = np.std(y, ddof=1) ** 2
    return winvar


def WinsorizedCorrelation(x, y, trimming_level=0.2):
    """
    Calculate the Winsorized correlation between two samples.

    Parameters
    ----------
    x : array-like
        The first input sample.
    y : array-like
        The second input sample.
    trimming_level : float, optional
        The trimming level (default is 0.2).

    Returns
    -------
    dict
        A dictionary containing the Winsorized correlation, covariance, p-value, sample size, and test statistic.
    """
    sample_size = len(x)
    x_sorted = np.sort(x)
    y_sorted = np.sort(y)
    trimming_size = int(np.floor(trimming_level * sample_size)) + 1
    x_lower = x_sorted[trimming_size - 1]
    x_upper = x_sorted[sample_size - trimming_size]
    y_lower = y_sorted[trimming_size - 1]
    y_upper = y_sorted[sample_size - trimming_size]
    x_winzsorized = np.clip(x, x_lower, x_upper)
    y_winzsorized = np.clip(y, y_lower, y_upper)
    winsorized_correlation = np.corrcoef(x_winzsorized, y_winzsorized)[0, 1]
    winsorized_covariance = np.cov(x_winzsorized, y_winzsorized)[0, 1]
    test_statistic = winsorized_correlation * np.sqrt(
        (sample_size - 2) / (1 - winsorized_correlation**2)
    )
    Number_of_trimmed_values = int(np.floor(trimming_level * sample_size))
    p_value = 2 * (
        1
        - t.cdf(np.abs(test_statistic), sample_size - 2 * Number_of_trimmed_values - 2)
    )
    return {
        "cor": winsorized_correlation,
        "cov": winsorized_covariance,
        "p.value": p_value,
        "n": sample_size,
        "test_statistic": test_statistic,
    }


class OneSampleTTest:
    """
    A class containing static methods for calculating the Common Language Effect Size (CLES) and other related statistics.

    This class includes the following static methods:
    - one_sample_from_t_score: Calculate the one-sample t-test results from a given t-score.
    - one_sample_from_params: Calculate the one-sample t-test results from given parameters.
    - one_sample_from_data: Calculate the one-sample t-test results from given sample data.
    - Robust_One_Sample: Calculate the robust one-sample t-test results.
    """

    @staticmethod
    def one_sample_from_t_score(params: dict) -> OneSampleCLESResults:
        """
        Calculate the one-sample t-test results from a given t-score.

        Parameters
        ----------
        params : dict
            A dictionary containing the following keys:
            - "t score" (float): The t-score value.
            - "Sample Size" (int): The size of the sample.
            - "Confidence Level" (float): The confidence level as a percentage (e.g., 95 for 95%).

        Returns
        -------
        OneSampleCLESResults
            An object containing all calculated statistics and results
        """
        # Get params
        t_score = params["t score"]
        sample_size = params["Sample Size"]
        confidence_level_percentages = params["Confidence Level"]

        # Create results object
        results = OneSampleCLESResults()

        # Calculation
        confidence_level = confidence_level_percentages / 100
        df = sample_size - 1
        p_value = min(float(t.sf((abs(t_score)), df) * 2), 0.99999)
        cohens_d = t_score / np.sqrt(
            df
        )  # This is Cohen's d and it is calculated based on the sample's standard deviation
        correction = math.exp(
            math.lgamma(df / 2)
            - math.log(math.sqrt(df / 2))
            - math.lgamma((df - 1) / 2)
        )
        hedges_g = correction * cohens_d
        cles_d = norm.cdf(cohens_d) * 100
        cles_g = norm.cdf(hedges_g) * 100

        (
            ci_lower_cohens_d_central,
            ci_upper_cohens_d_central,
            standrat_error_cohens_d,
        ) = calculate_central_ci_from_cohens_d_one_sample_t_test(
            cohens_d, sample_size, confidence_level
        )
        (
            ci_lower_hedges_g_central,
            ci_upper_hedges_g_central,
            standrat_error_hedges_g,
        ) = calculate_central_ci_from_cohens_d_one_sample_t_test(
            hedges_g, sample_size, confidence_level
        )
        ci_lower_cohens_d_pivotal, ci_upper_cohens_d_pivotal = pivotal_ci_t(
            t_score, df, sample_size, confidence_level
        )
        ci_lower_hedges_g_pivotal, ci_upper_hedges_g_pivotal = pivotal_ci_t(
            t_score, df, sample_size, confidence_level
        )
        ci_lower_cohens_d_ncp, ci_upper_cohens_d_ncp = ci_ncp_one_sample(
            cohens_d, sample_size, confidence_level
        )
        ci_lower_hedges_g_ncp, ci_upper_hedges_g_ncp = ci_ncp_one_sample(
            hedges_g, sample_size, confidence_level
        )

        # Set results
        results.cles_d = np.around(cles_d, 4)
        results.cles_g = np.around(cles_g, 4)
        results.t_score = np.around(t_score, 4)
        results.df = np.around(df, 4)
        results.p_value = np.around(p_value, 4)

        results.lower_central_ci_cld = np.around(
            norm.cdf(ci_lower_cohens_d_central) * 100, 4
        )
        results.upper_central_ci_cld = np.around(
            norm.cdf(ci_upper_cohens_d_central) * 100, 4
        )
        results.lower_central_ci_clg = np.around(
            norm.cdf(ci_lower_hedges_g_central) * 100, 4
        )
        results.upper_central_ci_clg = np.around(
            norm.cdf(ci_upper_hedges_g_central) * 100, 4
        )

        results.lower_ncp_ci_cld = np.around(norm.cdf(ci_lower_cohens_d_ncp) * 100, 4)
        results.upper_ncp_ci_cld = np.around(norm.cdf(ci_upper_cohens_d_ncp) * 100, 4)
        results.lower_ncp_ci_clg = np.around(norm.cdf(ci_lower_hedges_g_ncp) * 100, 4)
        results.upper_ncp_ci_clg = np.around(norm.cdf(ci_upper_hedges_g_ncp) * 100, 4)

        results.lower_pivotal_ci_cld = np.around(
            norm.cdf(ci_lower_cohens_d_pivotal) * 100, 4
        )
        results.upper_pivotal_ci_cld = np.around(
            norm.cdf(ci_upper_cohens_d_pivotal) * 100, 4
        )
        results.lower_pivotal_ci_clg = np.around(
            norm.cdf(ci_lower_hedges_g_pivotal * correction) * 100, 4
        )
        results.upper_pivotal_ci_clg = np.around(
            norm.cdf(ci_upper_hedges_g_pivotal * correction) * 100, 4
        )
        formatted_p_value = (
            "{:.3f}".format(p_value).lstrip("0")
            if p_value >= 0.001
            else "\033[3mp\033[0m < .001"
        )
        results.statistical_line_cld = "\033[3mt\033[0m({}) = {:.3f}, {}{}, CLd = {:.3f}, {}% CI(pivotal) [{:.3f}, {:.3f}]".format(
            df,
            t_score,
            "\033[3mp = \033[0m" if p_value >= 0.001 else "",
            formatted_p_value,
            cles_d,
            confidence_level_percentages,
            np.around(norm.cdf(ci_lower_cohens_d_pivotal) * 100, 3),
            np.around(norm.cdf(ci_upper_cohens_d_pivotal) * 100, 3),
        )
        results.statistical_line_clg = "\033[3mt\033[0m({}) = {:.3f}, {}{}, CLg = {:.3f}, {}% CI(pivotal) [{:.3f}, {:.3f}]".format(
            df,
            t_score,
            "\033[3mp = \033[0m" if p_value >= 0.001 else "",
            formatted_p_value,
            cles_g,
            confidence_level_percentages,
            np.around(norm.cdf(ci_lower_hedges_g_pivotal * correction) * 100, 3),
            np.around(norm.cdf(ci_upper_hedges_g_pivotal * correction) * 100, 3),
        )

        return results

    @staticmethod
    def one_sample_from_params(params: dict) -> OneSampleCLESResults:
        """
        Calculate the one-sample t-test results from given parameters.

        Parameters
        ----------
        params : dict
            A dictionary containing the following keys:
            - "Population Mean" (float): The population mean value.
            - "Mean Sample" (float): The mean of the sample.
            - "Standard Deviation Sample" (float): The standard deviation of the sample.
            - "Sample Size" (int): The size of the sample.
            - "Confidence Level" (float): The confidence level as a percentage (e.g., 95 for 95%).

        Returns
        -------
        OneSampleCLESResults
            An object containing all calculated statistics and results
        """
        # Set params
        population_mean = params["Population Mean"]
        sample_mean = params["Mean Sample"]
        sample_sd = params["Standard Deviation Sample"]
        sample_size = params["Sample Size"]
        confidence_level_percentages = params["Confidence Level"]

        # Create results object
        results = OneSampleCLESResults()

        # Calculation
        confidence_level = confidence_level_percentages / 100
        df = sample_size - 1
        standard_error = sample_sd / np.sqrt(
            df
        )  # This is the standard error of mean's estimate in one sample t-test
        t_score = (
            sample_mean - population_mean
        ) / standard_error  # This is the t score in the test which is used to calculate the p-value
        cohens_d = (
            sample_mean - population_mean
        ) / sample_sd  # This is the effect size for one sample t-test Cohen's d
        correction = math.exp(
            math.lgamma(df / 2)
            - math.log(math.sqrt(df / 2))
            - math.lgamma((df - 1) / 2)
        )
        hedges_g = cohens_d * correction  # This is the actual corrected effect size
        cles_d = norm.cdf(cohens_d) * 100
        cles_g = norm.cdf(hedges_g) * 100
        p_value = min(float(t.sf((abs(t_score)), df) * 2), 0.99999)
        (
            ci_lower_cohens_d_central,
            ci_upper_cohens_d_central,
            standrat_error_cohens_d,
        ) = calculate_central_ci_from_cohens_d_one_sample_t_test(
            cohens_d, sample_size, confidence_level
        )
        (
            ci_lower_hedges_g_central,
            ci_upper_hedges_g_central,
            standrat_error_hedges_g,
        ) = calculate_central_ci_from_cohens_d_one_sample_t_test(
            hedges_g, sample_size, confidence_level
        )
        ci_lower_cohens_d_pivotal, ci_upper_cohens_d_pivotal = pivotal_ci_t(
            t_score, df, sample_size, confidence_level
        )
        ci_lower_hedges_g_pivotal, ci_upper_hedges_g_pivotal = pivotal_ci_t(
            t_score, df, sample_size, confidence_level
        )
        ci_lower_cohens_d_ncp, ci_upper_cohens_d_ncp = ci_ncp_one_sample(
            cohens_d, sample_size, confidence_level
        )
        ci_lower_hedges_g_ncp, ci_upper_hedges_g_ncp = ci_ncp_one_sample(
            hedges_g, sample_size, confidence_level
        )

        # Set results
        results.cles_d = np.around(cles_d, 4)
        results.cles_g = np.around(cles_g, 4)
        results.t_score = np.around(t_score, 4)
        results.df = np.around(df, 4)
        results.p_value = np.around(p_value, 4)

        results.lower_central_ci_cld = np.around(
            norm.cdf(ci_lower_cohens_d_central) * 100, 4
        )
        results.upper_central_ci_cld = np.around(
            norm.cdf(ci_upper_cohens_d_central) * 100, 4
        )
        results.lower_central_ci_clg = np.around(
            norm.cdf(ci_lower_hedges_g_central) * 100, 4
        )
        results.upper_central_ci_clg = np.around(
            norm.cdf(ci_upper_hedges_g_central) * 100, 4
        )

        results.lower_ncp_ci_cld = np.around(norm.cdf(ci_lower_cohens_d_ncp) * 100, 4)
        results.upper_ncp_ci_cld = np.around(norm.cdf(ci_upper_cohens_d_ncp) * 100, 4)
        results.lower_ncp_ci_clg = np.around(norm.cdf(ci_lower_hedges_g_ncp) * 100, 4)
        results.upper_ncp_ci_clg = np.around(norm.cdf(ci_upper_hedges_g_ncp) * 100, 4)

        results.lower_pivotal_ci_cld = np.around(
            norm.cdf(ci_lower_cohens_d_pivotal) * 100, 4
        )
        results.upper_pivotal_ci_cld = np.around(
            norm.cdf(ci_upper_cohens_d_pivotal) * 100, 4
        )
        results.lower_pivotal_ci_clg = np.around(
            norm.cdf(ci_lower_hedges_g_pivotal * correction) * 100, 4
        )
        results.upper_pivotal_ci_clg = np.around(
            norm.cdf(ci_upper_hedges_g_pivotal * correction) * 100, 4
        )
        formatted_p_value = (
            "{:.3f}".format(p_value).lstrip("0")
            if p_value >= 0.001
            else "\033[3mp\033[0m < .001"
        )
        results.statistical_line_cld = "\033[3mt\033[0m({}) = {:.3f}, {} {}, CLd = {:.3f}, {}% CI(pivotal) [{:.3f}, {:.3f}]".format(
            int(df),
            t_score,
            "\033[3mp = \033[0m" if p_value >= 0.001 else "",
            formatted_p_value,
            cles_d,
            confidence_level_percentages,
            np.around(norm.cdf(ci_lower_cohens_d_pivotal) * 100, 3),
            np.around(norm.cdf(ci_upper_cohens_d_pivotal) * 100, 3),
        )
        results.statistical_line_clg = "\033[3mt\033[0m({}) = {:.3f}, {} {}, CLg = {:.3f}, {}% CI(pivotal) [{:.3f}, {:.3f}]".format(
            int(df),
            t_score,
            "\033[3mp = \033[0m" if p_value >= 0.001 else "",
            formatted_p_value,
            cles_g,
            confidence_level_percentages,
            np.around(norm.cdf(ci_lower_hedges_g_pivotal * correction) * 100, 3),
            np.around(norm.cdf(ci_upper_hedges_g_pivotal * correction) * 100, 3),
        )

        return results

    @staticmethod
    def one_sample_from_data(params: dict) -> OneSampleCLESResults:
        """
        Calculate the one-sample t-test results from given sample data.

        Parameters
        ----------
        params : dict
            A dictionary containing the following keys:
            - "column 1" (array-like): The sample data.
            - "Population's Mean" (float): The population mean value.
            - "Confidence Level" (float): The confidence level as a percentage (e.g., 95 for 95%).

        Returns
        -------
        dict
            A dictionary containing the calculated results:
            - "Mcgraw & Wong, Common Language Effect Size (CLd)" (float): The Common Language Effect Size (CLd).
            - "Mcgraw & Wong, Unbiased Common Language Effect Size (CLg)" (float): The Unbiased Common Language Effect Size (CLg).
            - "t-score" (float): The t-score value.
            - "degrees of freedom" (float): The degrees of freedom.
            - "p-value" (float): The p-value.
            - "Lower Central CI's CLd" (float): Lower bound of the central confidence interval for CLd.
            - "Upper Central CI's CLd" (float): Upper bound of the central confidence interval for CLd.
            - "Lower Central CI's CLg" (float): Lower bound of the central confidence interval for CLg.
            - "Upper Central CI's CLg" (float): Upper bound of the central confidence interval for CLg.
            - "Lower Non-Central CI's CLd" (float): Lower bound of the non-central confidence interval for CLd.
            - "Upper Non-Central CI's CLd" (float): Upper bound of the non-central confidence interval for CLd.
            - "Lower Non-Central CI's CLg" (float): Lower bound of the non-central confidence interval for CLg.
            - "Upper Non-Central CI's CLg" (float): Upper bound of the non-central confidence interval for CLg.
            - "Lower Pivotal CI's CLd" (float): Lower bound of the pivotal confidence interval for CLd.
            - "Upper Pivotal CI's CLd" (float): Upper bound of the pivotal confidence interval for CLd.
            - "Lower Pivotal CI's CLg" (float): Lower bound of the pivotal confidence interval for CLg.
            - "Upper Pivotal CI's CLg" (float): Upper bound of the pivotal confidence interval for CLg.
            - "Statistical Line CLd" (str): A formatted string with the statistical results for CLd.
            - "Statistical Line CLg" (str): A formatted string with the statistical results for CLg.
        """
        # Set params
        column_1 = params["column 1"]
        population_mean = params[
            "Population's Mean"
        ]  # Default should be 0 if not mentioned
        confidence_level_percentages = params["Confidence Level"]

        # Calculation
        confidence_level = confidence_level_percentages / 100
        sample_mean = np.mean(column_1)
        sample_sd = np.std(column_1, ddof=1)
        sample_size = len(column_1)
        df = sample_size - 1  # Degrees of freedom one sample t-test
        standard_error = sample_sd / np.sqrt(
            df
        )  # This is the standard error of mean's estimate in one sample t-test
        t_score = (
            sample_mean - population_mean
        ) / standard_error  # This is the t score in the test which is used to calculate the p-value
        cohens_d = (
            sample_mean - population_mean
        ) / sample_sd  # This is the effect size for one sample t-test Cohen's d
        correction = math.exp(
            math.lgamma(df / 2)
            - math.log(math.sqrt(df / 2))
            - math.lgamma((df - 1) / 2)
        )
        hedges_g = cohens_d * correction  # This is the actual corrected effect size
        p_value = min(float(t.sf((abs(t_score)), df) * 2), 0.99999)
        (
            ci_lower_cohens_d_central,
            ci_upper_cohens_d_central,
            standrat_error_cohens_d,
        ) = calculate_central_ci_from_cohens_d_one_sample_t_test(
            cohens_d, sample_size, confidence_level
        )
        (
            ci_lower_hedges_g_central,
            ci_upper_hedges_g_central,
            standrat_error_hedges_g,
        ) = calculate_central_ci_from_cohens_d_one_sample_t_test(
            hedges_g, sample_size, confidence_level
        )
        ci_lower_cohens_d_pivotal, ci_upper_cohens_d_pivotal = pivotal_ci_t(
            t_score, df, sample_size, confidence_level
        )
        ci_lower_hedges_g_pivotal, ci_upper_hedges_g_pivotal = pivotal_ci_t(
            t_score, df, sample_size, confidence_level
        )
        ci_lower_cohens_d_ncp, ci_upper_cohens_d_ncp = ci_ncp_one_sample(
            cohens_d, sample_size, confidence_level
        )
        ci_lower_hedges_g_ncp, ci_upper_hedges_g_ncp = ci_ncp_one_sample(
            hedges_g, sample_size, confidence_level
        )
        cles_d = norm.cdf(cohens_d) * 100
        cles_g = norm.cdf(hedges_g) * 100

        # Non Parametric Common Language Effect Sizes
        # Consider Adding Common language effect size to test the probability of a sample value to be larger than the median in the population...(as in matlab mes package)

        # Create results object
        results = OneSampleCLESResults()

        # Set results
        results.cles_d = np.around(cles_d, 4)
        results.cles_g = np.around(cles_g, 4)
        results.t_score = np.around(t_score, 4)
        results.df = np.around(df, 4)
        results.p_value = np.around(p_value, 4)

        results.lower_central_ci_cld = np.around(
            norm.cdf(ci_lower_cohens_d_central) * 100, 4
        )
        results.upper_central_ci_cld = np.around(
            norm.cdf(ci_upper_cohens_d_central) * 100, 4
        )
        results.lower_central_ci_clg = np.around(
            norm.cdf(ci_lower_hedges_g_central) * 100, 4
        )
        results.upper_central_ci_clg = np.around(
            norm.cdf(ci_upper_hedges_g_central) * 100, 4
        )

        results.lower_ncp_ci_cld = np.around(norm.cdf(ci_lower_cohens_d_ncp) * 100, 4)
        results.upper_ncp_ci_cld = np.around(norm.cdf(ci_upper_cohens_d_ncp) * 100, 4)
        results.lower_ncp_ci_clg = np.around(norm.cdf(ci_lower_hedges_g_ncp) * 100, 4)
        results.upper_ncp_ci_clg = np.around(norm.cdf(ci_upper_hedges_g_ncp) * 100, 4)

        results.lower_pivotal_ci_cld = np.around(
            norm.cdf(ci_lower_cohens_d_pivotal) * 100, 4
        )
        results.upper_pivotal_ci_cld = np.around(
            norm.cdf(ci_upper_cohens_d_pivotal) * 100, 4
        )
        results.lower_pivotal_ci_clg = np.around(
            norm.cdf(ci_lower_hedges_g_pivotal * correction) * 100, 4
        )
        results.upper_pivotal_ci_clg = np.around(
            norm.cdf(ci_upper_hedges_g_pivotal * correction) * 100, 4
        )

        return results

    @staticmethod
    def robust_one_sample(params: dict) -> OneSampleCLESResults:
        """
        Calculate the robust one-sample t-test results.

        Parameters
        ----------
        params : dict
            A dictionary containing the following keys:
            - "column 1" (array-like): The sample data.
            - "Trimming Level" (float): The trimming level.
            - "Population Mean" (float): The population mean value.
            - "Number of Bootstrap Samples" (int): The number of bootstrap samples.
            - "Confidence Level" (float): The confidence level as a percentage (e.g., 95 for 95%).
        Returns
        -------
        OneSampleCLESResults
            An object containing all calculated statistics and results:
            - robust_effect_size_akp (float): The robust effect size (AKP)
            - lower_ci_robust_akp (float): Lower bound of the confidence interval for the robust effect size (AKP)
            - upper_ci_robust_akp (float): Upper bound of the confidence interval for the robust effect size (AKP)
            - trimmed_mean (float): The trimmed mean of the sample
            - winsorized_sd (float): The Winsorized standard deviation of the sample
            - yuens_t (float): The Yuen's T statistic
            - robust_df (float): The degrees of freedom
            - robust_p_value (float): The p-value
            - mean_difference (float): The difference between means
            - standard_error (float): The standard error
        """
        # Set Parameters
        column_1 = params["column 1"]
        trimming_level = params["Trimming Level"]  # The default should be 0.2
        population_mean = params["Population Mean"]  # The default should be 0.2
        reps = params["Number of Bootstrap Samples"]
        confidence_level_percentages = params["Confidence Level"]

        # Calculation
        confidence_level = confidence_level_percentages / 100
        sample_size = len(column_1)
        difference = np.array(column_1) - population_mean
        correction = np.sqrt(
            area_under_function(
                density, norm.ppf(trimming_level), norm.ppf(1 - trimming_level)
            )
            + 2 * (norm.ppf(trimming_level) ** 2) * trimming_level
        )
        trimmed_mean_1 = trim_mean(column_1, trimming_level)
        winsorized_standard_deviation_1 = np.sqrt(WinsorizedVariance(column_1))

        # Algina, Penfield, Kesselman robust effect size (AKP)
        standardizer = np.sqrt(WinsorizedVariance(difference, trimming_level))
        trimmed_mean = trim_mean(difference, trimming_level)
        akp_effect_size = correction * (trimmed_mean - population_mean) / standardizer

        # Confidence Intervals for AKP effect size using Bootstrapping
        bootstrap_difference = []
        for _ in range(reps):
            # Generate bootstrap samples
            difference_bootstrap = np.random.choice(
                difference, len(difference), replace=True
            )
            bootstrap_difference.append(difference_bootstrap)

        trimmed_means_of_bootstrap = trim_mean(
            bootstrap_difference, trimming_level, axis=1
        )
        standardizers_of_bootstrap = np.sqrt(
            [
                WinsorizedVariance(array, trimming_level)
                for array in bootstrap_difference
            ]
        )
        akp_effect_size_bootstrap = (
            correction
            * (trimmed_means_of_bootstrap - population_mean)
            / standardizers_of_bootstrap
        )
        lower_ci_akp_boot = np.percentile(
            akp_effect_size_bootstrap,
            ((1 - confidence_level) - ((1 - confidence_level) / 2)) * 100,
        )
        upper_ci_akp_boot = np.percentile(
            akp_effect_size_bootstrap,
            ((confidence_level) + ((1 - confidence_level) / 2)) * 100,
        )

        # Yuen Test Statistics
        non_winsorized_sample_size = len(column_1) - 2 * np.floor(
            trimming_level * len(column_1)
        )
        df = non_winsorized_sample_size - 1
        yuen_standard_error = winsorized_standard_deviation_1 / (
            (1 - 2 * trimming_level) * np.sqrt(len(column_1))
        )
        difference_trimmed_means = trimmed_mean_1 - population_mean
        yuen_statistic = difference_trimmed_means / yuen_standard_error
        yuen_p_value = 2 * (1 - t.cdf(np.abs(yuen_statistic), df))

        # Set results
        results = OneSampleCLESResults()

        results.robust_effect_size_akp = round(akp_effect_size, 4)
        results.lower_ci_robust_akp = round(lower_ci_akp_boot, 4)
        results.upper_ci_robust_akp = round(upper_ci_akp_boot, 4)

        # Descriptive Statistics
        results.trimmed_mean = round(trimmed_mean_1, 4)
        results.winsorized_sd = round(winsorized_standard_deviation_1, 4)

        # Inferential Statistics
        results.yuens_t = round(yuen_statistic, 4)
        results.robust_df = round(df, 4)
        results.robust_p_value = np.around(yuen_p_value, 4)
        results.mean_difference = round(difference_trimmed_means, 4)
        results.standard_error = round(yuen_standard_error, 4)

        return results

    # Things to consider
    # 1. Consider adding a robust one-sample measures here
    # 2. Consider adding the z value transformed option
    # 3. Consider adding the mes matlab package for one sample CLES



"""
This module provides functionality for calculating the Aparametric effect size using the sign test for one sample.

Classes:
    AparametricOneSample: A class containing static methods for calculating the Aparametric effect size.

Methods:
    ApermetricEffectSizeOneSample: Calculate the Aparametric effect size using the sign test for one sample.
"""

import math
from dataclasses import dataclass
from typing import Optional
import numpy as np
from scipy.stats import norm, rankdata, median_abs_deviation


# Create results class
@dataclass
class OneSampleAparametricResults:
    """
    A class to store results from one-sample aparametric statistical tests.

    This class contains attributes to store various statistical measures including:
    - General summary statistics (sample size, means, medians etc.)
    - Wilcoxon test statistics (ignoring ties)
    - Rank biserial correlations
    - Confidence intervals
    - Statistical lines in formatted output
    - Pratt test statistics (considering ties)
    """

    sample: Optional[int | float] = None
    sample_median: Optional[int | float] = None
    median_of_the_difference: Optional[int | float] = None
    median_of_absolute_deviation: Optional[int | float] = None
    sample_mean: Optional[int | float] = None
    sample_standard_deviation: Optional[int | float] = None
    number_of_pairs: Optional[int | float] = None
    number_of_pairs_with_a_sign: Optional[int | float] = None
    number_of_times_sample_is_larger: Optional[int | float] = None
    number_of_times_sample_is_smaller: Optional[int | float] = None
    number_of_ties: Optional[int | float] = None

    # Wilcoxon Statistics (Wilcoxon Method that Ignores ties)
    wilcoxon_method: str = ""
    _______________: str = ""
    sum_of_the_positive_ranks_without_ties: Optional[int | float] = None
    sum_of_the_negative_ranks_without_ties: Optional[int | float] = None

    # Wilcoxon Sign Rank Test Statistics (Wilcoxon)
    wilcoxon_mean_w_without_ties: Optional[int | float] = None
    wilcoxon_standard_deviation: Optional[int | float] = None
    wilcoxon_z: Optional[int | float] = None
    wilcoxon_z_with_normal_approximation_continuity_correction: Optional[
        int | float
    ] = None
    wilcoxon_p_value: Optional[int | float] = None
    wilcoxon_p_value_with_normal_approximation_continuity_correction: Optional[
        int | float
    ] = None

    # Rank Biserial Correlation
    matched_pairs_rank_biserial_correlation_ignoring_ties: Optional[int | float] = None
    z_based_rank_biserial_correlation_wilcoxon_method: Optional[int | float] = None
    z_based_corrected_rank_biserial_correlation_wilcoxon_method: Optional[
        int | float
    ] = None

    # Confidence Intervals
    standard_error_of_the_matched_pairs_rank_biserial_correlation_wilcoxon_method: (
        Optional[int | float]
    ) = (None)
    lower_ci_matched_pairs_rank_biserial_wilcoxon: Optional[int | float] = None
    upper_ci_matched_pairs_rank_biserial_wilcoxon: Optional[int | float] = None
    lower_ci_z_based_rank_biserial_wilcoxon: Optional[int | float] = None
    upper_ci_z_based_rank_biserial_wilcoxon: Optional[int | float] = None
    lower_ci_z_based_corrected_rank_biserial_wilcoxon: Optional[int | float] = None
    upper_ci_z_based_corrected_rank_biserial_wilcoxon: Optional[int | float] = None

    # Statistical Lines Wilcoxon Method
    statistical_line_wilcoxon: Optional[int | float] = None
    statistical_line_wilcoxon_corrected: Optional[int | float] = None
    statistical_line_wilcoxon_matched_pairs: Optional[int | float] = None

    pratt_method: str = ""
    sum_of_the_positive_ranks_with_ties: Optional[int | float] = None
    sum_of_the_negative_ranks_with_ties: Optional[int | float] = None

    pratt_meanw_considering_ties: Optional[int | float] = None
    pratt_standard_deviation: Optional[int | float] = None
    pratt_z: Optional[int | float] = None
    pratt_z_with_normal_approximation_continuity_correction: Optional[int | float] = (
        None
    )
    pratt_p_value: Optional[int | float] = None
    pratt_p_value_with_normal_approximation_continuity_correction: Optional[
        int | float
    ] = None

    # Rank Biserial Correlation
    matched_pairs_rank_biserial_correlation_considering_ties: Optional[int | float] = (
        None
    )
    z_based_rank_biserial_correlation_pratt_method: Optional[int | float] = None
    z_based_corrected_rank_biserial_correlation_pratt_method: Optional[int | float] = (
        None
    )

    # Confidence Intervals
    standard_error_of_the_matched_pairs_rank_biserial_correlation_pratt_method: (
        Optional[int | float]
    ) = (None)
    lower_ci_matched_pairs_rank_biserial_pratt: Optional[int | float] = None
    upper_ci_matched_pairs_rank_biserial_pratt: Optional[int | float] = None
    lower_ci_z_based_rank_biserial_pratt: Optional[int | float] = None
    upper_ci_z_based_rank_biserial_pratt: Optional[int | float] = None
    lower_ci_z_based_corrected_rank_biserial_pratt: Optional[int | float] = None
    upper_ci_z_based_corrected_rank_biserial_pratt: Optional[int | float] = None

    # Statistical Lines
    statistical_line_pratt: Optional[int | float] = None
    statistical_line_pratt_corrected: Optional[int | float] = None
    statistical_line_pratt_matched_pairs: Optional[int | float] = None


def apermetric_effect_size_one_sample(params: dict) -> OneSampleAparametricResults:
    """
    Calculate the Aparametric effect size using the sign test for one sample.

    Parameters:
    params (dict): A dictionary containing the following keys:
        - "Column 1": A numpy array of sample data.
        - "Population's Value": The population value to compare against.
        - "Confidence Level": The confidence level as a percentage.

    Returns:
    dict: A dictionary containing the results of the Aparametric effect size calculations.
    """
    # region
    # Set Parameters
    column_1 = params["Column 1"]
    population_value = params["Population's Value"]
    confidence_level_percentages = params["Confidence Level"]

    # Calculation
    confidence_level = confidence_level_percentages / 100

    # General Summary Statistics
    sample_median_1 = np.median(column_1)
    sample_mean_1 = np.mean(column_1)
    sample_standard_deviation_1 = float(np.std(column_1, ddof=1))
    difference = column_1 - population_value
    positive_n = difference[difference > 0].shape[
        0
    ]  # How many times sample is greater than population value
    negative_n = difference[difference < 0].shape[
        0
    ]  # How many times sample is lower than population value
    zero_n = difference[difference == 0].shape[0]  # Number of ties
    sample_size = len(difference)
    median_difference = np.median(difference)
    median_absolute_deviation = float(median_abs_deviation(difference))

    # Summary Statistics for the Wilcoxon Sign Rank Test not Considering ties
    difference_no_ties = difference[difference != 0]  # This line removes the ties
    ranked_no_ties = rankdata(abs(difference_no_ties))
    positive_sum_ranks_no_ties = ranked_no_ties[difference_no_ties > 0].sum()
    negative_sum_ranks_no_ties = ranked_no_ties[difference_no_ties < 0].sum()

    # Summary Statistics for the Wilcoxon Sign Rank Considering ties
    ranked_with_ties = rankdata(abs(difference))
    positive_sum_ranks_with_ties = ranked_with_ties[difference > 0].sum()
    negative_sum_ranks_with_ties = ranked_with_ties[difference < 0].sum()

    # Wilcoxon Sign Rank Test Statistics Non Considering Ties (Wilcoxon Method)
    mean_w_not_considering_ties = (
        positive_sum_ranks_no_ties + negative_sum_ranks_no_ties
    ) / 2
    sign_no_ties = np.where(
        difference_no_ties == 0, 0, (np.where(difference_no_ties < 0, -1, 1))
    )
    ranked_signs_no_ties = sign_no_ties * ranked_no_ties
    ranked_signs_no_ties = np.where(difference_no_ties == 0, 0, ranked_signs_no_ties)
    unadjusted_variance_wilcoxon = (
        len(difference_no_ties)
        * (len(difference_no_ties) + 1)
        * (2 * (len(difference_no_ties)) + 1)
    ) / 24
    var_adj_t = (ranked_signs_no_ties * ranked_signs_no_ties).sum()
    adjusted_variance_wilcoxon = (1 / 4) * var_adj_t

    # Calculate The Z score wilcox
    z_numerator_wilcoxon = positive_sum_ranks_no_ties - mean_w_not_considering_ties
    z_numerator_wilcoxon = np.where(
        z_numerator_wilcoxon < 0, z_numerator_wilcoxon + 0.5, z_numerator_wilcoxon
    )

    z_adjusted_wilcoxon = (z_numerator_wilcoxon) / np.sqrt(adjusted_variance_wilcoxon)
    z_adjusted_normal_approximation_wilcoxon = (z_numerator_wilcoxon - 0.5) / np.sqrt(
        adjusted_variance_wilcoxon
    )
    z_unadjusted_wilcoxon = (z_numerator_wilcoxon) / np.sqrt(
        unadjusted_variance_wilcoxon
    )
    z_unadjusted_normal_approximation_wilcoxon = (z_numerator_wilcoxon - 0.5) / np.sqrt(
        unadjusted_variance_wilcoxon
    )
    p_value_adjusted_wilcoxon = min(
        float(norm.sf((abs(z_adjusted_wilcoxon))) * 2), 0.99999
    )
    p_value_adjusted_normal_approximation_wilcoxon = min(
        float(norm.sf((abs(z_adjusted_normal_approximation_wilcoxon))) * 2),
        0.99999,
    )
    p_value_unadjusted_wilcoxon = min(
        float(norm.sf((abs(z_unadjusted_wilcoxon))) * 2), 0.99999
    )
    p_value_unadjusted_normal_approximation_wilcoxon = min(
        float(norm.sf((abs(z_unadjusted_normal_approximation_wilcoxon))) * 2),
        0.99999,
    )

    # Wilcoxon Sign Rank Test Statistics Considering Ties (Pratt Method)
    mean_w_considering_ties = (
        positive_sum_ranks_with_ties + negative_sum_ranks_with_ties
    ) / 2
    sign_with_ties = np.where(difference == 0, 0, (np.where(difference < 0, -1, 1)))
    ranked_signs_with_ties = sign_with_ties * ranked_with_ties
    ranked_signs_with_ties = np.where(difference == 0, 0, ranked_signs_with_ties)
    var_adj_t_with_ties = (ranked_signs_with_ties * ranked_signs_with_ties).sum()
    adjusted_variance_pratt = (1 / 4) * var_adj_t_with_ties

    z_numerator_pratt = positive_sum_ranks_with_ties - mean_w_considering_ties
    z_numerator_pratt = np.where(
        z_numerator_pratt < 0, z_numerator_pratt + 0.5, z_numerator_pratt
    )

    z_adjusted_pratt = (z_numerator_pratt) / np.sqrt(adjusted_variance_pratt)
    z_adjusted_normal_approximation_pratt = (z_numerator_pratt - 0.5) / np.sqrt(
        adjusted_variance_pratt
    )
    p_value_adjusted_pratt = min(float(norm.sf((abs(z_adjusted_pratt))) * 2), 0.99999)
    p_value_adjusted_normal_approximation_pratt = min(
        float(norm.sf((abs(z_adjusted_normal_approximation_pratt))) * 2), 0.99999
    )

    # Matched Pairs Rank Biserial Correlation
    matched_pairs_rank_biserial_correlation_ignoring_ties = min(
        (positive_sum_ranks_no_ties - negative_sum_ranks_no_ties)
        / np.sum(ranked_no_ties),
        0.99999999,
    )  # This is the match paired rank biserial correlation using kerby formula that is not considering ties (Kerby, 2014)
    matched_pairs_rank_biserial_correlation_considering_ties = min(
        (positive_sum_ranks_with_ties - negative_sum_ranks_with_ties)
        / np.sum(ranked_with_ties),
        0.999999999,
    )  # this is the Kerby 2014 Formula - (With ties one can apply either Kerby or King Minium Formulae but not cureton - King's Formula is the most safe)

    # Z-based Rank Biserial Correlation (Note that since the Wilcoxon method is ignoring ties the sample size should actually be the number of the non tied pairs)
    z_based_rank_biserial_correlation_no_ties = z_adjusted_wilcoxon / np.sqrt(
        len(ranked_no_ties)
    )
    z_based_rank_biserial_correlation_corrected_no_ties = (
        z_adjusted_normal_approximation_wilcoxon / np.sqrt(len(ranked_no_ties))
    )
    z_based_rank_biserial_correlation_with_ties = z_adjusted_pratt / np.sqrt(
        sample_size
    )
    z_based_rank_biserial_correlation_corrected_with_ties = (
        z_adjusted_normal_approximation_pratt / np.sqrt(sample_size)
    )

    # Confidence Intervals
    standard_error_match_pairs_rank_biserial_correlation_no_ties = np.sqrt(
        (
            (
                2 * (len(ranked_no_ties)) ** 3
                + 3 * (len(ranked_no_ties)) ** 2
                + (len(ranked_no_ties))
            )
            / 6
        )
        / (((len(ranked_no_ties)) ** 2 + (len(ranked_no_ties)) / 2))
    )
    standard_error_match_pairs_rank_biserial_correlation_with_ties = np.sqrt(
        ((2 * sample_size**3 + 3 * sample_size**2 + sample_size) / 6)
        / ((sample_size**2 + sample_size) / 2)
    )
    z_critical_value = norm.ppf((1 - confidence_level) + ((confidence_level) / 2))

    lower_ci_matched_pairs_wilcoxon = max(
        math.tanh(
            math.atanh(matched_pairs_rank_biserial_correlation_ignoring_ties)
            - z_critical_value
            * standard_error_match_pairs_rank_biserial_correlation_no_ties
        ),
        -1,
    )
    upper_ci_matched_pairs_wilcoxon = min(
        math.tanh(
            math.atanh(matched_pairs_rank_biserial_correlation_ignoring_ties)
            + z_critical_value
            * standard_error_match_pairs_rank_biserial_correlation_no_ties
        ),
        1,
    )
    lower_ci_z_based_wilcoxon = max(
        math.tanh(
            math.atanh(z_based_rank_biserial_correlation_no_ties)
            - z_critical_value
            * standard_error_match_pairs_rank_biserial_correlation_no_ties
        ),
        -1,
    )
    upper_ci_z_based_wilcoxon = min(
        math.tanh(
            math.atanh(z_based_rank_biserial_correlation_no_ties)
            + z_critical_value
            * standard_error_match_pairs_rank_biserial_correlation_no_ties
        ),
        1,
    )
    lower_ci_z_based_corrected_wilcoxon = max(
        math.tanh(
            math.atanh(z_based_rank_biserial_correlation_corrected_no_ties)
            - z_critical_value
            * standard_error_match_pairs_rank_biserial_correlation_no_ties
        ),
        -1,
    )
    upper_ci_z_based_corrected_wilcoxon = min(
        math.tanh(
            math.atanh(z_based_rank_biserial_correlation_corrected_no_ties)
            + z_critical_value
            * standard_error_match_pairs_rank_biserial_correlation_no_ties
        ),
        1,
    )

    lower_ci_matched_pairs_pratt = max(
        math.tanh(
            math.atanh(matched_pairs_rank_biserial_correlation_considering_ties)
            - z_critical_value
            * standard_error_match_pairs_rank_biserial_correlation_with_ties
        ),
        -1,
    )
    upper_ci_matched_pairs_pratt = min(
        math.tanh(
            math.atanh(matched_pairs_rank_biserial_correlation_considering_ties)
            + z_critical_value
            * standard_error_match_pairs_rank_biserial_correlation_with_ties
        ),
        1,
    )
    lower_ci_z_based_pratt = max(
        math.tanh(
            math.atanh(z_based_rank_biserial_correlation_with_ties)
            - z_critical_value
            * standard_error_match_pairs_rank_biserial_correlation_with_ties
        ),
        -1,
    )
    upper_ci_z_based_pratt = min(
        math.tanh(
            math.atanh(z_based_rank_biserial_correlation_with_ties)
            + z_critical_value
            * standard_error_match_pairs_rank_biserial_correlation_with_ties
        ),
        1,
    )
    lower_ci_z_based_corrected_pratt = max(
        math.tanh(
            math.atanh(z_based_rank_biserial_correlation_corrected_with_ties)
            - z_critical_value
            * standard_error_match_pairs_rank_biserial_correlation_with_ties
        ),
        -1,
    )
    upper_ci_z_based_corrected_pratt = min(
        math.tanh(
            math.atanh(z_based_rank_biserial_correlation_corrected_with_ties)
            + z_critical_value
            * standard_error_match_pairs_rank_biserial_correlation_with_ties
        ),
        1,
    )
    # endregion

    results = OneSampleAparametricResults()

    # General Summary Statistics
    results.sample = positive_sum_ranks_with_ties
    results.sample_median = round(sample_median_1, 4)
    results.median_of_the_difference = median_difference
    results.median_of_absolute_deviation = median_absolute_deviation
    results.sample_mean = round(sample_mean_1, 4)
    results.sample_standard_deviation = round(sample_standard_deviation_1, 4)
    results.number_of_pairs = sample_size
    results.number_of_pairs_with_a_sign = len(ranked_no_ties)
    results.number_of_times_sample_is_larger = positive_n
    results.number_of_times_sample_is_smaller = negative_n
    results.number_of_ties = zero_n

    # Wilcoxon Statistics (Wilcoxon Method that Ignores ties)
    results.sum_of_the_positive_ranks_without_ties = round(
        positive_sum_ranks_no_ties, 4
    )
    results.sum_of_the_negative_ranks_without_ties = round(
        negative_sum_ranks_no_ties, 4
    )

    # Wilcoxon Sign Rank Test Statistics (Wilcoxon)
    results.wilcoxon_mean_w_without_ties = mean_w_not_considering_ties
    results.wilcoxon_standard_deviation = np.sqrt(adjusted_variance_wilcoxon)
    results.wilcoxon_z = z_adjusted_wilcoxon
    results.wilcoxon_z_with_normal_approximation_continuity_correction = (
        z_adjusted_normal_approximation_wilcoxon
    )
    results.wilcoxon_p_value = p_value_adjusted_wilcoxon
    results.wilcoxon_p_value_with_normal_approximation_continuity_correction = (
        p_value_adjusted_normal_approximation_wilcoxon
    )

    # Rank Biserial Correlation
    results.matched_pairs_rank_biserial_correlation_ignoring_ties = round(
        matched_pairs_rank_biserial_correlation_ignoring_ties, 5
    )
    results.z_based_rank_biserial_correlation_wilcoxon_method = round(
        z_based_rank_biserial_correlation_no_ties, 5
    )
    results.z_based_corrected_rank_biserial_correlation_wilcoxon_method = round(
        z_based_rank_biserial_correlation_corrected_no_ties, 5
    )

    # Confidence Intervals
    results.standard_error_of_the_matched_pairs_rank_biserial_correlation_wilcoxon_method = round(
        standard_error_match_pairs_rank_biserial_correlation_no_ties, 4
    )
    results.lower_ci_matched_pairs_rank_biserial_wilcoxon = round(
        lower_ci_matched_pairs_wilcoxon, 5
    )
    results.upper_ci_matched_pairs_rank_biserial_wilcoxon = round(
        upper_ci_matched_pairs_wilcoxon, 5
    )
    results.lower_ci_z_based_rank_biserial_wilcoxon = round(
        lower_ci_z_based_wilcoxon, 5
    )
    results.upper_ci_z_based_rank_biserial_wilcoxon = round(
        upper_ci_z_based_wilcoxon, 5
    )
    results.lower_ci_z_based_corrected_rank_biserial_wilcoxon = round(
        lower_ci_z_based_corrected_wilcoxon, 5
    )
    results.upper_ci_z_based_corrected_rank_biserial_wilcoxon = round(
        upper_ci_z_based_corrected_wilcoxon, 5
    )

    results.sum_of_the_positive_ranks_with_ties = round(positive_sum_ranks_with_ties, 4)
    results.sum_of_the_negative_ranks_with_ties = round(negative_sum_ranks_with_ties, 4)

    results.pratt_meanw_considering_ties = mean_w_considering_ties
    results.pratt_standard_deviation = np.sqrt(adjusted_variance_pratt)
    results.pratt_z = z_adjusted_pratt
    results.pratt_z_with_normal_approximation_continuity_correction = (
        z_adjusted_normal_approximation_pratt
    )
    results.pratt_p_value = p_value_adjusted_pratt
    results.pratt_p_value_with_normal_approximation_continuity_correction = (
        p_value_adjusted_normal_approximation_pratt
    )

    # Rank Biserial Correlation
    results.matched_pairs_rank_biserial_correlation_considering_ties = round(
        matched_pairs_rank_biserial_correlation_considering_ties, 5
    )
    results.z_based_rank_biserial_correlation_pratt_method = round(
        z_based_rank_biserial_correlation_with_ties, 5
    )
    results.z_based_corrected_rank_biserial_correlation_pratt_method = round(
        z_based_rank_biserial_correlation_corrected_with_ties, 5
    )

    # Confidence Intervals
    results.standard_error_of_the_matched_pairs_rank_biserial_correlation_pratt_method = round(
        standard_error_match_pairs_rank_biserial_correlation_with_ties, 4
    )
    results.lower_ci_matched_pairs_rank_biserial_pratt = round(
        lower_ci_matched_pairs_pratt, 5
    )
    results.upper_ci_matched_pairs_rank_biserial_pratt = round(
        upper_ci_matched_pairs_pratt, 5
    )
    results.lower_ci_z_based_rank_biserial_pratt = round(lower_ci_z_based_pratt, 5)
    results.upper_ci_z_based_rank_biserial_pratt = round(upper_ci_z_based_pratt, 5)
    results.lower_ci_z_based_corrected_rank_biserial_pratt = round(
        lower_ci_z_based_corrected_pratt, 5
    )
    results.upper_ci_z_based_corrected_rank_biserial_pratt = round(
        upper_ci_z_based_corrected_pratt, 5
    )

    return results


# Things to Consider
# 1. Consider adding other CI's for example metsamuuronen method for sommers delta (which in the case of two groups equals the rank biserial correlation)
# 2. Test if the matched pairs version is also equal to Sommers delta and cliffs delta (dependent version)
# 3. For convenience change the confidence levels to percentages and not decimals



"""
This module implements robust statistical tests for two paired samples.
It includes methods for calculating robust effect sizes, confidence intervals,
and inferential statistics using Yuen's t-test and winsorized Pearson correlation.
It also provides a data class to store the results of these tests.
"""

from dataclasses import dataclass
from typing import Optional, Callable
import numpy as np
from scipy.stats import t, norm, trim_mean
from ...utils import interfaces
from ...utils import res


@dataclass
class TwoPairedRobustResults:
    """
    A data class to store the results of two paired robust tests.
        It includes samples, inferential statistics, robust effect sizes,
    """

    sample1: Optional[res.Sample] = None
    sample2: Optional[res.Sample] = None
    inferential: Optional[res.InferentialStatistics] = None
    robust_akp: Optional[res.RobustAKP] = None
    robust_explanatory: Optional[res.RobustExplanatory] = None
    yuen_robust_t: Optional[float] = None
    winsorized_pearson_correlation: Optional[float] = None
    winsorized_pearson_correlation_p_value: Optional[float] = None


class TwoPairedRobustTests(interfaces.AbstractTest):
    """
    A class to perform robust statistical tests for two paired samples.
    It includes methods for calculating robust effect sizes, confidence intervals,
    and inferential statistics using Yuen's t-test and winsorized Pearson correlation.

        Methods:
                - from_score: Not implemented.
                - from_parameters: Not implemented.
                - from_data: Calculates robust statistics from two paired samples.
                - density: Calculates the density function for a given input.
                - area_under_function: Computes the area under a given function between two points.
                - winsorized_variance: Computes the winsorized variance of a sample.
                - winsorized_correlation: Computes the winsorized Pearson correlation between two samples.
    """

    @staticmethod
    def from_score() -> TwoPairedRobustResults:
        """
        A static method to create results from a score.
        This method is not implemented and will raise a NotImplementedError.
        Raises:
                NotImplementedError: This method is not implemented for TwoPairedRobustResults.

        """
        raise NotImplementedError(
            "from_score method is not implemented for TwoPairedRobust."
        )

    @staticmethod
    def from_parameters() -> TwoPairedRobustResults:
        """
        A static method to create results from parameters.
            This method is not implemented and will raise a NotImplementedError.
        Raises:
                    NotImplementedError: This method is not implemented for TwoPairedRobustResults.
        """
        raise NotImplementedError(
            "from_parameters method is not implemented for TwoPairedRobust."
        )

    def from_data(
        self,
        columns: list,
        reps: int,
        confidence_level: float,
        trimming_level: float = 0.2,
        population_difference: float = 0.2,
    ) -> TwoPairedRobustResults:
        """
        A method to calculate robust statistics from two paired samples.
        """

        column_1 = columns[0]
        column_2 = columns[1]

        difference = np.array(column_1) - np.array(column_2)
        sample_size = len(column_1)
        correction = np.sqrt(
            float(
                self.area_under_function(
                    self.density,
                    float(norm.ppf(trimming_level)),
                    float(norm.ppf(1 - trimming_level)),
                )
            )
            + 2 * (norm.ppf(trimming_level) ** 2) * trimming_level
        )
        trimmed_mean_1 = trim_mean(column_1, trimming_level)
        trimmed_mean_2 = trim_mean(column_2, trimming_level)
        winsorized_standard_deviation_1 = np.sqrt(self.winsorized_variance(column_1))
        winsorized_standard_deviation_2 = np.sqrt(self.winsorized_variance(column_2))
        winsorized = self.winsorized_correlation(column_1, column_2, trimming_level)
        winsorized_correlation = winsorized["cor"]
        winsorized_correlation_p_value = winsorized["p.value"]

        standardizer = np.sqrt(self.winsorized_variance(difference, trimming_level))
        trimmed_mean = trim_mean(difference, trimming_level)
        akp_effect_size = (
            correction * (trimmed_mean - population_difference) / standardizer
        )

        bootstrap_samples_difference = []
        for _ in range(reps):
            difference_bootstrap = np.random.choice(
                difference, len(difference), replace=True
            )
            bootstrap_samples_difference.append(difference_bootstrap)

        trimmed_means_of_bootstrap = trim_mean(
            bootstrap_samples_difference, trimming_level, axis=1
        )
        standardizer_of_bootstrap = np.sqrt(
            [
                self.winsorized_variance(array, trimming_level)
                for array in bootstrap_samples_difference
            ]
        )
        akp_effect_size_bootstrap = (
            correction
            * (trimmed_means_of_bootstrap - population_difference)
            / standardizer_of_bootstrap
        )
        lower_ci_akp_boot = np.percentile(
            akp_effect_size_bootstrap,
            ((1 - confidence_level) - ((1 - confidence_level) / 2)) * 100,
        )
        upper_ci_akp_boot = np.percentile(
            akp_effect_size_bootstrap,
            ((confidence_level) + ((1 - confidence_level) / 2)) * 100,
        )

        sort_values = np.concatenate((column_1, column_2))
        variance_between_trimmed_means = (
            np.std(np.array([trimmed_mean_1, trimmed_mean_2]), ddof=1)
        ) ** 2
        winsorized_variance_value = self.winsorized_variance(
            sort_values, trimming_level
        )
        explained_variance = variance_between_trimmed_means / (
            winsorized_variance_value / correction**2
        )
        explanatory_power_effect_size = min(float(np.sqrt(explained_variance)), 1)

        bootstrap_samples_x = []
        bootstrap_samples_y = []
        for _ in range(reps):
            sample_1_bootstrap = np.random.choice(column_1, len(column_1), replace=True)
            sample_2_bootstrap = np.random.choice(column_2, len(column_2), replace=True)
            bootstrap_samples_x.append(sample_1_bootstrap)
            bootstrap_samples_y.append(sample_2_bootstrap)

        concatenated_samples = [
            np.concatenate((x, y))
            for x, y in zip(bootstrap_samples_x, bootstrap_samples_y)
        ]
        trimmed_means_of_bootstrap_sample_1 = np.array(
            (trim_mean(bootstrap_samples_x, trimming_level, axis=1))
        )
        trimmed_means_of_bootstrap_sample_2 = np.array(
            (trim_mean(bootstrap_samples_y, trimming_level, axis=1))
        )
        variance_between_trimmed_means_bootstrap = [
            (np.std(np.array([x, y]), ddof=1)) ** 2
            for x, y in zip(
                trimmed_means_of_bootstrap_sample_1, trimmed_means_of_bootstrap_sample_2
            )
        ]
        winsorized_variances_bootstrap = [
            self.winsorized_variance(arr, trimming_level)
            for arr in concatenated_samples
        ]
        explained_variance_bootstrapping = np.array(
            variance_between_trimmed_means_bootstrap
            / (winsorized_variances_bootstrap / correction**2)
        )
        explanatory_power_effect_size_bootstrap = [
            array**0.5 for array in explained_variance_bootstrapping
        ]
        lower_ci_e_pow_boot = np.percentile(
            explanatory_power_effect_size_bootstrap,
            ((1 - confidence_level) - ((1 - confidence_level) / 2)) * 100,
        )
        upper_ci_e_pow_boot = np.percentile(
            explanatory_power_effect_size_bootstrap,
            ((confidence_level) + ((1 - confidence_level) / 2)) * 100,
        )

        q1 = (len(column_1) - 1) * winsorized_standard_deviation_1**2
        q2 = (len(column_2) - 1) * winsorized_standard_deviation_2**2
        q3 = (len(column_1) - 1) * winsorized["cov"]
        non_winsorized_sample_size = len(column_1) - 2 * np.floor(
            trimming_level * len(column_1)
        )
        df = non_winsorized_sample_size - 1
        yuen_standard_error = np.sqrt(
            (q1 + q2 - 2 * q3)
            / (non_winsorized_sample_size * (non_winsorized_sample_size - 1))
        )
        difference_trimmed_means = trimmed_mean_1 - trimmed_mean_2
        yuen_statistic = difference_trimmed_means / yuen_standard_error
        yuen_p_value = 2 * (1 - t.cdf(np.abs(yuen_statistic), df))

        sample1 = res.Sample(
            mean=round(trimmed_mean_1, 4),
            standard_deviation=round(winsorized_standard_deviation_1, 4),
            size=round(sample_size, 4),
        )
        sample2 = res.Sample(
            mean=round(trimmed_mean_2, 4),
            standard_deviation=round(winsorized_standard_deviation_2, 4),
            size=round(sample_size, 4),
        )

        inferential = res.InferentialStatistics(
            p_value=float(np.around(yuen_p_value, 4)),
            score=round(yuen_statistic, 4),
        )
        inferential.standard_error = round(yuen_standard_error, 4)
        inferential.degrees_of_freedom = round(df, 4)
        inferential.means_difference = round(difference_trimmed_means, 4)

        robust_akp = res.RobustAKP(
            value=round(akp_effect_size, 4),
            ci_lower=float(round(lower_ci_akp_boot, 4)),
            ci_upper=float(round(upper_ci_akp_boot, 4)),
            standard_error=round(yuen_standard_error, 4),
        )
        robust_explanatory = res.RobustExplanatory(
            value=round(explanatory_power_effect_size, 4),
            ci_lower=float(round(lower_ci_e_pow_boot, 4)),
            ci_upper=round(min(float(upper_ci_e_pow_boot), 1.0), 4),
            standard_error=round(yuen_standard_error, 4),
        )

        results = TwoPairedRobustResults()
        results.sample1 = sample1
        results.sample2 = sample2
        results.inferential = inferential
        results.robust_akp = robust_akp
        results.robust_explanatory = robust_explanatory
        results.yuen_robust_t = round(yuen_statistic, 4)
        results.winsorized_pearson_correlation = round(winsorized_correlation, 4)
        results.winsorized_pearson_correlation_p_value = round(
            winsorized_correlation_p_value, 4
        )

        return results

    def density(self, x):
        x = np.array(x)
        return x**2 * norm.pdf(x)

    def area_under_function(
        self,
        f: Callable,
        a,
        b,
        *args,
        function_a=None,
        function_b=None,
        limit=10,
        eps=1e-5,
    ):
        if function_a is None:
            function_a = f(a, *args)
        if function_b is None:
            function_b = f(b, *args)
        midpoint = (a + b) / 2
        f_midpoint = f(midpoint, *args)
        area_trapezoidal = ((function_a + function_b) * (b - a)) / 2
        area_simpson = ((function_a + 4 * f_midpoint + function_b) * (b - a)) / 6
        if abs(area_trapezoidal - area_simpson) < eps or limit == 0:
            return area_simpson
        return self.area_under_function(
            f,
            a,
            midpoint,
            *args,
            function_a=function_a,
            function_b=f_midpoint,
            limit=limit - 1,
            eps=eps,
        ) + self.area_under_function(
            f,
            midpoint,
            b,
            *args,
            function_a=f_midpoint,
            function_b=function_b,
            limit=limit - 1,
            eps=eps,
        )

    def winsorized_variance(self, x, trimming_level=0.2):
        y = np.sort(x)
        n = len(x)
        ibot = int(np.floor(trimming_level * n)) + 1
        itop = n - ibot + 1
        xbot = y[ibot - 1]
        xtop = y[itop - 1]
        y = np.where(y <= xbot, xbot, y)
        y = np.where(y >= xtop, xtop, y)
        winvar = np.std(y, ddof=1) ** 2
        return winvar

    def winsorized_correlation(self, x, y, trimming_level=0.2):
        sample_size = len(x)
        x_sorted = np.sort(x)
        y_sorted = np.sort(y)
        trimming_size = int(np.floor(trimming_level * sample_size)) + 1
        x_lower = x_sorted[trimming_size - 1]
        x_upper = x_sorted[sample_size - trimming_size]
        y_lower = y_sorted[trimming_size - 1]
        y_upper = y_sorted[sample_size - trimming_size]
        x_winsorized = np.clip(x, x_lower, x_upper)
        y_winsorized = np.clip(y, y_lower, y_upper)
        winsorized_correlation = np.corrcoef(x_winsorized, y_winsorized)[0, 1]
        winsorized_covariance = np.cov(x_winsorized, y_winsorized)[0, 1]
        test_statistic = winsorized_correlation * np.sqrt(
            (sample_size - 2) / (1 - winsorized_correlation**2)
        )
        number_of_trimmed_values = int(np.floor(trimming_level * sample_size))
        p_value = 2 * (
            1
            - t.cdf(
                np.abs(test_statistic), sample_size - 2 * number_of_trimmed_values - 2
            )
        )
        return {
            "cor": winsorized_correlation,
            "cov": winsorized_covariance,
            "p.value": p_value,
            "n": sample_size,
            "test_statistic": test_statistic,
        }



"""
This module contains functions and classes for performing one-sample t-tests and
calculating various statistics such as Cohen's d, Hedges' g, t-score, p-value,
confidence intervals, and standard errors.

The module includes the following functions:
- pivotal_ci_t: Calculate the Pivotal confidence intervals for a one-sample t-test.
- calculate_central_ci_one_sample_t_test: Calculate the central confidence intervals
for the effect size in a one-sample t-test.
- CI_NCP_one_Sample: Calculate the Non-Central Parameter (NCP) confidence intervals
for a one-sample t-test.

The module also includes the following class:
- One_Sample_ttest: A class containing static methods for performing one-sample
t-tests from t-score and parameters.
"""

import math
from typing import Optional
from dataclasses import dataclass
import numpy as np
from scipy.stats import norm, nct, t
from ...utils import interfaces
from ...utils import res


@dataclass
class OneSampleTResults:
    """
    A class to store results from one-sample t-tests.
    """

    # Effect sizes
    cohens_d: Optional[res.CohenD] = None
    hedges_g: Optional[res.HedgesG] = None
    # Test statistics
    t_score: Optional[float] = None
    degrees_of_freedom: Optional[int | float] = None
    p_value: Optional[float] = None
    standard_error: Optional[float | int] = None
    sample_mean: Optional[float | int] = None
    population_mean: Optional[float | int] = None
    means_difference: Optional[float | int] = None
    sample_size: Optional[float | int] = None
    sample_sd: Optional[float | int] = None


def pivotal_ci_t(t_score, df, sample_size, confidence_level):
    """
    Calculate the Pivotal confidence intervals for a one-sample t-test.

    Parameters
    ----------
    t_score : float
        The t-score value.
    df : int
        Degrees of freedom.
    sample_size : int
        The size of the sample.
    confidence_level : float
        The confidence level as a decimal (e.g., 0.95 for 95%).

    Returns
    -------
    tuple
        A tuple containing:
        - lower_ci (float): Lower bound of the confidence interval.
        - upper_ci (float): Upper bound of the confidence interval.
    """
    is_negative = False
    if t_score < 0:
        is_negative = True
        t_score = abs(t_score)
    upper_limit = 1 - (1 - confidence_level) / 2
    lower_limit = (1 - confidence_level) / 2

    lower_criterion = [-t_score, t_score / 2, t_score]
    upper_criterion = [t_score, 2 * t_score, 3 * t_score]

    while nct.cdf(t_score, df, lower_criterion[0]) < upper_limit:
        lower_criterion = [
            lower_criterion[0] - t_score,
            lower_criterion[0],
            lower_criterion[2],
        ]

    while nct.cdf(t_score, df, upper_criterion[0]) < lower_limit:
        if nct.cdf(t_score, df) < lower_limit:
            lower_ci = [0, nct.cdf(t_score, df)]
            upper_criterion = [
                upper_criterion[0] / 4,
                upper_criterion[0],
                upper_criterion[2],
            ]

    while nct.cdf(t_score, df, upper_criterion[2]) > lower_limit:
        upper_criterion = [
            upper_criterion[0],
            upper_criterion[2],
            upper_criterion[2] + t_score,
        ]

    lower_ci = 0.0
    diff_lower = 1
    while diff_lower > 0.00001:
        if nct.cdf(t_score, df, lower_criterion[1]) < upper_limit:
            lower_criterion = [
                lower_criterion[0],
                (lower_criterion[0] + lower_criterion[1]) / 2,
                lower_criterion[1],
            ]
        else:
            lower_criterion = [
                lower_criterion[1],
                (lower_criterion[1] + lower_criterion[2]) / 2,
                lower_criterion[2],
            ]
        diff_lower = abs(nct.cdf(t_score, df, lower_criterion[1]) - upper_limit)
        lower_ci = lower_criterion[1] / (np.sqrt(sample_size))

    upper_ci = 0.0
    diff_upper = 1
    while diff_upper > 0.00001:
        if nct.cdf(t_score, df, upper_criterion[1]) < lower_limit:
            upper_criterion = [
                upper_criterion[0],
                (upper_criterion[0] + upper_criterion[1]) / 2,
                upper_criterion[1],
            ]
        else:
            upper_criterion = [
                upper_criterion[1],
                (upper_criterion[1] + upper_criterion[2]) / 2,
                upper_criterion[2],
            ]
        diff_upper = abs(nct.cdf(t_score, df, upper_criterion[1]) - lower_limit)
        upper_ci = upper_criterion[1] / (np.sqrt(sample_size))
    if is_negative:
        return -upper_ci, -lower_ci
    else:
        return lower_ci, upper_ci


def calculate_central_ci_one_sample_t_test(effect_size, sample_size, confidence_level):
    """
    Calculate the central confidence intervals for the effect size in a one-sample t-test.

    Parameters
    ----------
    effect_size : float
        The calculated effect size (Cohen's d).
    sample_size : int
        The size of the sample.
    confidence_level : float
        The confidence level as a decimal (e.g., 0.95 for 95%).

    Returns
    -------
    tuple
        A tuple containing:
        - ci_lower (float): Lower bound of the confidence interval.
        - ci_upper (float): Upper bound of the confidence interval.
        - Standard_error_effect_size_True (float): Standard error of the effect size (True).
        - Standard_error_effect_size_Morris (float): Standard error of the effect size (Morris).
        - Standard_error_effect_size_Hedges (float): Standard error of the effect size (Hedges).
        - Standard_error_effect_size_Hedges_Olkin (float): Standard error of the effect size (Hedges_Olkin).
        - Standard_error_effect_size_MLE (float): Standard error of the effect size (MLE).
        - Standard_error_effect_size_Large_N (float): Standard error of the effect size (Large N).
        - Standard_error_effect_size_Small_N (float): Standard error of the effect size (Small N).
    """
    df = sample_size - 1
    correction_factor = math.exp(
        math.lgamma(df / 2) - math.log(math.sqrt(df / 2)) - math.lgamma((df - 1) / 2)
    )
    standard_error_effect_size_true = np.sqrt(
        (
            (df / (df - 2)) * (1 / sample_size) * (1 + effect_size**2 * sample_size)
            - (effect_size**2 / correction_factor**2)
        )
    )
    standard_error_effect_size_morris = np.sqrt(
        (df / (df - 2)) * (1 / sample_size) * (1 + effect_size**2 * sample_size)
        - (effect_size**2 / (1 - (3 / (4 * (df - 1) - 1))) ** 2)
    )
    standard_error_effect_size_hedges = np.sqrt(
        (1 / sample_size) + effect_size**2 / (2 * df)
    )
    standard_error_effect_size_hedges_olkin = np.sqrt(
        (1 / sample_size) + effect_size**2 / (2 * sample_size)
    )
    standard_error_effect_size_mle = np.sqrt(
        standard_error_effect_size_hedges * ((df + 2) / df)
    )
    standard_error_effect_size_large_n = np.sqrt(
        1 / sample_size * (1 + effect_size**2 / 8)
    )
    standard_error_effect_size_small_n = np.sqrt(
        standard_error_effect_size_large_n * ((df + 1) / (df - 1))
    )
    z_critical_value = norm.ppf(confidence_level + ((1 - confidence_level) / 2))
    ci_lower, ci_upper = (
        effect_size - standard_error_effect_size_true * z_critical_value,
        effect_size + standard_error_effect_size_true * z_critical_value,
    )
    return (
        ci_lower,
        ci_upper,
        standard_error_effect_size_true,
        standard_error_effect_size_morris,
        standard_error_effect_size_hedges,
        standard_error_effect_size_hedges_olkin,
        standard_error_effect_size_mle,
        standard_error_effect_size_large_n,
        standard_error_effect_size_small_n,
    )


def ci_ncp_one_sample(effect_size, sample_size, confidence_level):
    """
    Calculate the Non-Central Parameter (NCP) confidence intervals for a one-sample t-test.

    Parameters
    ----------
    effect_size : float
        The calculated effect size (Cohen's d).
    sample_size : int
        The size of the sample.
    confidence_level : float
        The confidence level as a decimal (e.g., 0.95 for 95%).

    Returns
    -------
    tuple
        A tuple containing:
        - CI_NCP_low (float): Lower bound of the NCP confidence interval.
        - CI_NCP_High (float): Upper bound of the NCP confidence interval.
    """
    NCP_value = effect_size * math.sqrt(sample_size)
    CI_NCP_low = (
        (
            nct.ppf(
                1 / 2 - confidence_level / 2,
                (sample_size - 1),
                loc=0,
                scale=1,
                nc=NCP_value,
            )
        )
        / NCP_value
        * effect_size
    )
    CI_NCP_High = (
        (
            nct.ppf(
                1 / 2 + confidence_level / 2,
                (sample_size - 1),
                loc=0,
                scale=1,
                nc=NCP_value,
            )
        )
        / NCP_value
        * effect_size
    )
    return CI_NCP_low, CI_NCP_High


class OneSampleTTest(interfaces.AbstractTest):
    """
    A class to perform one-sample t-tests and calculate various statistics.
    This class provides methods to calculate T-test results from a t-score,
    from sample parameters, and from sample data.
    """

    @staticmethod
    def from_score(
        t_score: float, sample_size: float, confidence_level=0.95
    ) -> OneSampleTResults:
        """
        Calculate the one-sample t-test results from a given t-score.
        """

        # Calculation
        df = sample_size - 1
        p_value = min(float(t.sf((abs(t_score)), df) * 2), 0.99999)
        cohens_d = t_score / np.sqrt(
            df
        )  # This is Cohen's d and it is calculated based on the sample's standard deviation
        correction = math.exp(
            math.lgamma(df / 2)
            - math.log(math.sqrt(df / 2))
            - math.lgamma((df - 1) / 2)
        )
        hedges_g = correction * cohens_d
        (
            ci_lower_cohens_d_central,
            ci_upper_cohens_d_central,
            standard_error_cohens_d_true,
            standard_error_cohens_d_morris,
            standard_error_cohens_d_hedges,
            standard_error_cohens_d_hedges_olkin,
            standard_error_cohens_d_mle,
            standard_error_cohens_d_large_n,
            standard_error_cohens_d_small_n,
        ) = calculate_central_ci_one_sample_t_test(
            cohens_d, sample_size, confidence_level
        )
        (
            ci_lower_hedges_g_central,
            ci_upper_hedges_g_central,
            standard_error_hedges_g_true,
            standard_error_hedges_g_morris,
            standard_error_hedges_g_hedges,
            standard_error_hedges_g_hedges_olkin,
            standard_error_hedges_g_mle,
            standard_error_hedges_g_large_n,
            standard_error_hedges_g_small_n,
        ) = calculate_central_ci_one_sample_t_test(
            hedges_g, sample_size, confidence_level
        )
        ci_lower_cohens_d_pivotal, ci_upper_cohens_d_pivotal = pivotal_ci_t(
            t_score, df, sample_size, confidence_level
        )
        ci_lower_hedges_g_pivotal, ci_upper_hedges_g_pivotal = pivotal_ci_t(
            t_score, df, sample_size, confidence_level
        )
        ci_lower_cohens_d_ncp, ci_upper_cohens_d_ncp = ci_ncp_one_sample(
            cohens_d, sample_size, confidence_level
        )
        ci_lower_hedges_g_ncp, ci_upper_hedges_g_ncp = ci_ncp_one_sample(
            hedges_g, sample_size, confidence_level
        )

        cohens_d = res.CohenD(
            cohens_d,
            ci_lower_cohens_d_central,
            ci_upper_cohens_d_central,
            standard_error_cohens_d_true,
        )

        cohens_d.standardizer = correction
        cohens_d.update_non_central_ci(ci_lower_cohens_d_ncp, ci_upper_cohens_d_ncp)
        cohens_d.update_pivotal_ci(ci_lower_cohens_d_pivotal, ci_upper_cohens_d_pivotal)

        cohens_d_approximated = res.ApproximatedStandardError(
            standard_error_cohens_d_true,
            standard_error_cohens_d_morris,
            standard_error_cohens_d_hedges,
            standard_error_cohens_d_hedges_olkin,
            standard_error_cohens_d_mle,
            standard_error_cohens_d_large_n,
            standard_error_cohens_d_small_n,
        )

        cohens_d.approximated_standard_error = cohens_d_approximated

        hedges_g = res.HedgesG(
            hedges_g,
            ci_lower_hedges_g_central,
            ci_upper_hedges_g_central,
            standard_error_hedges_g_true,
        )

        hedges_g.update_non_central_ci(ci_lower_hedges_g_ncp, ci_upper_hedges_g_ncp)
        hedges_g.update_pivotal_ci(ci_lower_hedges_g_pivotal, ci_upper_hedges_g_pivotal)

        hedges_g_approximated = res.ApproximatedStandardError(
            standard_error_hedges_g_true,
            standard_error_hedges_g_morris,
            standard_error_hedges_g_hedges,
            standard_error_hedges_g_hedges_olkin,
            standard_error_hedges_g_mle,
            standard_error_hedges_g_large_n,
            standard_error_hedges_g_small_n,
        )
        hedges_g.approximated_standard_error = hedges_g_approximated

        results = OneSampleTResults()
        results.t_score = t_score
        results.degrees_of_freedom = df
        results.p_value = p_value
        results.cohens_d = cohens_d
        results.hedges_g = hedges_g

        return results

    @staticmethod
    def from_parameters(
        population_mean: float,
        sample_mean: float,
        sample_sd: float,
        sample_size: float,
        confidence_level: float = 0.95,
    ) -> OneSampleTResults:
        """
        Calculate the one-sample t-test results from given parameters.
        """
        df = sample_size - 1
        standard_error = sample_sd / np.sqrt(
            df
        )  # This is the standrt error of mean's estimate in o ne samaple t-test
        t_score = (
            sample_mean - population_mean
        ) / standard_error  # This is the t score in the test which is used to calculate the p-value
        cohens_d = (
            sample_mean - population_mean
        ) / sample_sd  # This is the effect size for one sample t-test Cohen's d
        correction = math.exp(
            math.lgamma(df / 2)
            - math.log(math.sqrt(df / 2))
            - math.lgamma((df - 1) / 2)
        )
        hedges_g = cohens_d * correction  # This is the actual corrected effect size
        p_value = min(float(t.sf((abs(t_score)), df) * 2), 0.99999)
        (
            ci_lower_cohens_d_central,
            ci_upper_cohens_d_central,
            standard_error_cohens_d_true,
            standard_error_cohens_d_morris,
            standard_error_cohens_d_hedges,
            standard_error_cohens_d_hedges_olkin,
            standard_error_cohens_d_mle,
            standard_error_cohens_d_large_n,
            standard_error_cohens_d_small_n,
        ) = calculate_central_ci_one_sample_t_test(
            cohens_d, sample_size, confidence_level
        )
        (
            ci_lower_hedges_g_central,
            ci_upper_hedges_g_central,
            standard_error_hedges_g_true,
            standard_error_hedges_g_morris,
            standard_error_hedges_g_hedges,
            standard_error_hedges_g_hedges_olkin,
            standard_error_hedges_g_mle,
            standard_error_hedges_g_large_n,
            standard_error_hedges_g_small_n,
        ) = calculate_central_ci_one_sample_t_test(
            hedges_g, sample_size, confidence_level
        )
        ci_lower_cohens_d_pivotal, ci_upper_cohens_d_pivotal = pivotal_ci_t(
            t_score, df, sample_size, confidence_level
        )
        ci_lower_hedges_g_pivotal, ci_upper_hedges_g_pivotal = pivotal_ci_t(
            t_score, df, sample_size, confidence_level
        )
        ci_lower_cohens_d_ncp, ci_upper_cohens_d_ncp = ci_ncp_one_sample(
            cohens_d, sample_size, confidence_level
        )
        ci_lower_hedges_g_ncp, ci_upper_hedges_g_ncp = ci_ncp_one_sample(
            hedges_g, sample_size, confidence_level
        )

        cohens_d = res.CohenD(
            cohens_d,
            ci_lower_cohens_d_central,
            ci_upper_cohens_d_central,
            standard_error_cohens_d_true,
        )
        cohens_d.standardizer = correction
        cohens_d.update_non_central_ci(
            float(ci_lower_cohens_d_ncp), float(ci_upper_cohens_d_ncp)
        )
        cohens_d.update_pivotal_ci(ci_lower_cohens_d_pivotal, ci_upper_cohens_d_pivotal)
        cohens_d_approximated = res.ApproximatedStandardError(
            standard_error_cohens_d_true,
            standard_error_cohens_d_morris,
            standard_error_cohens_d_hedges,
            standard_error_cohens_d_hedges_olkin,
            standard_error_cohens_d_mle,
            standard_error_cohens_d_large_n,
            standard_error_cohens_d_small_n,
        )
        cohens_d.approximated_standard_error = cohens_d_approximated

        hedges_g = res.HedgesG(
            hedges_g,
            ci_lower_hedges_g_central,
            ci_upper_hedges_g_central,
            standard_error_hedges_g_true,
        )

        hedges_g.update_non_central_ci(
            float(ci_lower_hedges_g_ncp), float(ci_upper_hedges_g_ncp)
        )
        hedges_g.update_pivotal_ci(ci_lower_hedges_g_pivotal, ci_upper_hedges_g_pivotal)
        hedges_g_approximated = res.ApproximatedStandardError(
            standard_error_hedges_g_true,
            standard_error_hedges_g_morris,
            standard_error_hedges_g_hedges,
            standard_error_hedges_g_hedges_olkin,
            standard_error_hedges_g_mle,
            standard_error_hedges_g_large_n,
            standard_error_hedges_g_small_n,
        )
        hedges_g.approximated_standard_error = hedges_g_approximated

        # Create results object
        results = OneSampleTResults()

        results.t_score = t_score
        results.degrees_of_freedom = df
        results.p_value = p_value
        results.cohens_d = cohens_d
        results.hedges_g = hedges_g

        # Assign values to the results object
        results.sample_size = sample_size
        results.population_mean = population_mean
        results.sample_mean = sample_mean
        results.sample_sd = sample_sd
        results.standard_error = standard_error
        results.means_difference = sample_mean - population_mean

        return results

    @staticmethod
    def from_data() -> OneSampleTResults:
        """
        Calculate the one-sample t-test results from sample data.
        This method is not implemented yet.
        """
        raise NotImplementedError("This method is not implemented yet.")



"""
This module provides utility functions for the Calculator package in the stats project.
"""

from dataclasses import asdict, is_dataclass
from typing import Any
import numpy as np
from scipy.stats import norm

def convert_results_to_dict(dataclass_instance: Any) -> dict:
    """
    Converts a dataclass instance to a dictionary.

    Args:
        dataclass_instance (dataclass): An instance of a dataclass.

    Returns:
        dict: A dictionary representation of the dataclass instance.
    """
    if not (
        is_dataclass(dataclass_instance) and not isinstance(dataclass_instance, type)
    ):
        raise TypeError(
            f"Expected a dataclass instance, got: {type(dataclass_instance)}"
        )

    return asdict(dataclass_instance)


def central_ci_from_cohens_d(
    cohens_d: float, sample_size: float, confidence_level: float
) -> tuple[float, float, float]:
    """
    Calculate the confidence intervals and standard error for Cohen's d effect size in a
    one-sample Z-test.

    This function calculates the confidence intervals of the effect size (Cohen's d) for a
    one-sample Z-test or two dependent samples test using the Hedges and Olkin (1985)
    formula to estimate the standard error.

    Parameters
    ----------
    cohens_d : float
        The calculated Cohen's d effect size
    sample_size : float
        The size of the sample
    confidence_level : float
        The confidence level as a decimal (e.g., 0.95 for 95%)

    Returns
    -------
    tuple
        A tuple containing:
        - ci_lower (float): Lower bound of the confidence interval
        - ci_upper (float): Upper bound of the confidence interval
        - standard_error_es (float): Standard error of the effect size

    Notes
    -----
    Since the effect size in the population and its standard deviation are unknown,
    we estimate it based on the sample using the Hedges and Olkin (1985) formula
    to estimate the standard deviation of the effect size.
    """
    Standard_error_es = np.sqrt((1 / sample_size) + ((cohens_d**2 / (2 * sample_size))))
    z_critical_value = norm.ppf(confidence_level + ((1 - confidence_level) / 2))
    ci_lower, ci_upper = (
        cohens_d - Standard_error_es * z_critical_value,
        cohens_d + Standard_error_es * z_critical_value,
    )
    return ci_lower, ci_upper, Standard_error_es



"""
A module for calculating common language effect sizes for two paired means.
This module provides functionality to compute various effect sizes such as CLES, Hedges' g,
and Cliff's delta, as well as their confidence intervals and other related statistics.
It includes methods for calculating effect sizes from scores, parameters, and data,
and provides a results class to encapsulate the output of these calculations.
This module is part of the Esek library, which is designed for statistical analysis and effect size
calculations.
"""

from dataclasses import dataclass
import math
from typing import Optional
import numpy as np
from scipy.stats import norm, nct, beta, t
from ...utils import interfaces
from ...utils import res


@dataclass
class TwoPairedCommonLangResults:
    """
    A class to hold the results of common language effect size calculations for two paired means.
        This class contains various effect size results such as CLES, Hedges' g, probability
    """

    cles_cohen: Optional[res.CLES] = None
    cles_hedges: Optional[res.CLES] = None
    probability_of_superiority: Optional[res.ProbabilityOfSuperiority] = None
    vargha_delaney: Optional[res.VarghaDelaney] = None
    cliff_delta: Optional[res.CliffsDelta] = None
    non_param_u1: Optional[res.NonParametricU1] = None
    non_param_u3: Optional[res.NonParametricU3] = None
    kraemer_andrew_gamma: Optional[res.KraemerAndrewGamma] = None
    wilcox_musaka_q: Optional[res.WilcoxMusakaQ] = None


class TwoPairedCommonLangTests(interfaces.AbstractTest):
    """
    A class to perform common language effect size calculations for two paired means.
    This class provides methods to calculate effect sizes from t-scores, parameters, and data.
    It includes methods for calculating CLES, Hedges' g, Cliff's delta, and other
    effect sizes, along with their confidence intervals and related statistics.

    This class is part of the Esek library, which is designed for statistical analysis
    and effect size calculations.

    Attributes:
            None

    Methods:
            from_score() ->
                    TwoPairedCommonLangResults:
                    Calculates common language effect sizes from a t-score, sample size, and confidence level
            from_parameters() -> TwoPairedCommonLangResults:
                    Calculates common language effect sizes from sample means, standard deviations,
                    sample size, population mean difference, correlation, and confidence level.
            from_data() -> TwoPairedCommonLangResults:
                    Calculates common language effect sizes from data in two columns.
                    The columns should contain paired samples, and the method computes various effect sizes
                    including CLES, Hedges' g, Cliff's delta, and others, along with their confidence intervals.
                    The method also handles bootstrapping for non-parametric effect sizes.
            pivotal_ci() -> tuple:
                    Calculates pivotal confidence intervals for the effect sizes based on the t-score,
                    degrees of freedom, sample size, and confidence level.
            calculate_central_ci( -> tuple:
                    Calculates central confidence intervals for Cohen's d effect size based on the
                    effect size, sample size, and confidence level.

    """

    def from_score(
        self, t_score: float, sample_size: int, confidence_level: float
    ) -> TwoPairedCommonLangResults:
        """
        Calculates common language effect sizes from a t-score, sample size, and confidence level.
        """

        cohens_dz = t_score / np.sqrt(sample_size)
        cles_dz = norm.cdf(cohens_dz) * 100
        df = sample_size - 1
        correction = math.exp(
            math.lgamma(df / 2)
            - math.log(math.sqrt(df / 2))
            - math.lgamma((df - 1) / 2)
        )
        hedges_gz = correction * cohens_dz
        cles_gz = norm.cdf(hedges_gz) * 100
        p_value = min(float(t.sf((abs(t_score)), df) * 2), 0.99999)

        (
            ci_lower_cohens_dz_central,
            ci_upper_cohens_dz_central,
            standard_error_cohens_dz,
        ) = self.calculate_central_ci(cohens_dz, sample_size, confidence_level)
        (
            ci_lower_hedges_gz_central,
            ci_upper_hedges_gz_central,
            standard_error_hedges_gz,
        ) = self.calculate_central_ci(hedges_gz, sample_size, confidence_level)
        ci_lower_cohens_dz_pivotal, ci_upper_cohens_dz_pivotal = self.pivotal_ci(
            t_score, df, sample_size, confidence_level
        )
        ci_lower_hedges_gz_pivotal, ci_upper_hedges_gz_pivotal = self.pivotal_ci(
            t_score, df, sample_size, confidence_level
        )

        cles_cohen = res.CLES(
            method="cohen's dz",
            value=float(np.around(cles_dz, 4)),
            standard_error=standard_error_cohens_dz,
            ci_lower=float(np.around(norm.cdf(ci_lower_cohens_dz_central) * 100, 4)),
            ci_upper=float(np.around(norm.cdf(ci_upper_cohens_dz_central) * 100, 4)),
        )
        cles_hedges = res.CLES(
            method="hedges' gz",
            value=float(np.around(cles_gz, 4)),
            standard_error=standard_error_hedges_gz,
            ci_lower=float(np.around(norm.cdf(ci_lower_hedges_gz_central) * 100, 4)),
            ci_upper=float(np.around(norm.cdf(ci_upper_hedges_gz_central) * 100, 4)),
        )

        cles_cohen.update_pivotal_ci(
            pivotal_ci_lower=float(
                np.around(norm.cdf(ci_lower_cohens_dz_pivotal) * 100, 4)
            ),
            pivotal_ci_upper=float(
                np.around(norm.cdf(ci_upper_cohens_dz_pivotal) * 100, 4)
            ),
        )
        cles_hedges.update_pivotal_ci(
            pivotal_ci_lower=float(
                np.around(norm.cdf(ci_lower_hedges_gz_pivotal * correction) * 100, 4)
            ),
            pivotal_ci_upper=float(
                np.around(norm.cdf(ci_upper_hedges_gz_pivotal * correction) * 100, 4)
            ),
        )

        results = TwoPairedCommonLangResults()
        results.cles_cohen = cles_cohen
        results.cles_hedges = cles_hedges

        return results

    def from_parameters(
        self,
        sample_mean_1: float,
        sample_mean_2: float,
        sample_sd_1: float,
        sample_sd_2: float,
        sample_size: int,
        population_mean_diff: float,
        correlation: float,
        confidence_level: float,
    ) -> TwoPairedCommonLangResults:
        """
        Calculates common language effect sizes from sample means, standard deviations,
        """
        difference = sample_mean_1 - sample_mean_2
        mean_difference = np.mean(difference - population_mean_diff)

        standardizer_dz = np.sqrt(
            sample_sd_1**2
            + sample_sd_2**2
            - 2 * correlation * sample_sd_1 * sample_sd_2
        )
        cohens_dz = mean_difference / standardizer_dz
        cles_dz = norm.cdf(cohens_dz) * 100
        t_score = cohens_dz * np.sqrt(sample_size)
        df = sample_size - 1
        correction = math.exp(
            math.lgamma(df / 2)
            - math.log(math.sqrt(df / 2))
            - math.lgamma((df - 1) / 2)
        )
        hedges_gz = correction * cohens_dz
        cles_gz = norm.cdf(hedges_gz) * 100
        p_value = min(float(t.sf((abs(t_score)), df) * 2), 0.99999)

        (
            ci_lower_cohens_dz_central,
            ci_upper_cohens_dz_central,
            standard_error_cohens_dz,
        ) = self.calculate_central_ci(cohens_dz, sample_size, confidence_level)
        (
            ci_lower_hedges_gz_central,
            ci_upper_hedges_gz_central,
            standard_error_hedges_gz,
        ) = self.calculate_central_ci(hedges_gz, sample_size, confidence_level)
        ci_lower_cohens_dz_pivotal, ci_upper_cohens_dz_pivotal = self.pivotal_ci(
            t_score, df, sample_size, confidence_level
        )
        ci_lower_hedges_gz_pivotal, ci_upper_hedges_gz_pivotal = self.pivotal_ci(
            t_score, df, sample_size, confidence_level
        )

        cles_cohen = res.CLES(
            method="cohen's dz",
            value=float(np.around(cles_dz, 4)),
            standard_error=standard_error_cohens_dz,
            ci_lower=float(np.around(norm.cdf(ci_lower_cohens_dz_central) * 100, 4)),
            ci_upper=float(np.around(norm.cdf(ci_upper_cohens_dz_central) * 100, 4)),
        )
        cles_hedges = res.CLES(
            method="hedges' gz",
            value=float(np.around(cles_gz, 4)),
            standard_error=standard_error_hedges_gz,
            ci_lower=float(np.around(norm.cdf(ci_lower_hedges_gz_central) * 100, 4)),
            ci_upper=float(np.around(norm.cdf(ci_upper_hedges_gz_central) * 100, 4)),
        )
        cles_cohen.update_pivotal_ci(
            pivotal_ci_lower=float(
                np.around(norm.cdf(ci_lower_cohens_dz_pivotal) * 100, 4)
            ),
            pivotal_ci_upper=float(
                np.around(norm.cdf(ci_upper_cohens_dz_pivotal) * 100, 4)
            ),
        )
        cles_hedges.update_pivotal_ci(
            pivotal_ci_lower=float(
                np.around(norm.cdf(ci_lower_hedges_gz_pivotal * correction) * 100, 4)
            ),
            pivotal_ci_upper=float(
                np.around(norm.cdf(ci_upper_hedges_gz_pivotal * correction) * 100, 4)
            ),
        )

        results = TwoPairedCommonLangResults()
        results.cles_cohen = cles_cohen
        results.cles_hedges = cles_hedges

        return results

    def from_data(
        self, column: list, reps: int, confidence_level: float
    ) -> TwoPairedCommonLangResults:
        """
        Calculates common language effect sizes from data in two columns.
        """

        column_1 = column[0]
        column_2 = column[1]

        sample_1_mean = np.mean(column_1)
        sample_2_mean = np.mean(column_2)
        sample_1_median = np.median(column_1)
        sample_2_median = np.median(column_2)
        sample_1_standard_deviation = np.std(column_1, ddof=1)
        sample_2_standard_deviation = np.std(column_2, ddof=1)
        sample_size_1 = len(column_1)
        sample_size_2 = len(column_2)

        difference = column_1 - column_2
        mean_difference = np.mean(difference)
        standard_deviation_of_the_difference = np.std(difference, ddof=1)
        sample_size = len(difference)

        cohens_dz = mean_difference / standard_deviation_of_the_difference
        cles_dz = norm.cdf(cohens_dz) * 100
        t_score = cohens_dz * np.sqrt(sample_size)
        df = sample_size - 1
        correction = math.exp(
            math.lgamma(df / 2)
            - math.log(math.sqrt(df / 2))
            - math.lgamma((df - 1) / 2)
        )
        hedges_gz = correction * cohens_dz
        cles_gz = norm.cdf(hedges_gz) * 100
        p_value = min(float(t.sf((abs(t_score)), df) * 2), 0.99999)

        (
            ci_lower_cohens_dz_central,
            ci_upper_cohens_dz_central,
            standard_error_cohens_dz,
        ) = self.calculate_central_ci(cohens_dz, sample_size, confidence_level)
        (
            ci_lower_hedges_gz_central,
            ci_upper_hedges_gz_central,
            standard_error_hedges_gz,
        ) = self.calculate_central_ci(hedges_gz, sample_size, confidence_level)
        ci_lower_cohens_dz_pivotal, ci_upper_cohens_dz_pivotal = self.pivotal_ci(
            t_score, df, sample_size, confidence_level
        )
        ci_lower_hedges_gz_pivotal, ci_upper_hedges_gz_pivotal = self.pivotal_ci(
            t_score, df, sample_size, confidence_level
        )

        count_group1_larger = sum(np.where(difference > 0, 1, 0))
        count_group2_larger = sum(np.where(difference < 0, 1, 0))
        count_ties = sum(np.where(difference == 0, 1, 0))
        ps_dep = count_group1_larger / sample_size

        superiority_counts = np.where(
            column_1 > column_2, 1, np.where(column_1 < column_2, 0, 0.5)
        )
        vda_xy = sum(superiority_counts) / sample_size

        cliffs_delta = (
            sum(np.where(column_1 > column_2, 1, 0))
            - sum(np.where(column_2 > column_1, 1, 0))
        ) / sample_size

        ctag_square = norm.ppf(1 - confidence_level) ** 2
        ctag = norm.ppf(1 - confidence_level)
        A = ((count_group1_larger + 1) / (sample_size - count_group1_larger)) ** 2
        B = (
            81 * (count_group1_larger + 1) * (sample_size - count_group1_larger)
            - 9 * sample_size
            - 8
        )
        C = (
            -3
            * ctag
            * np.sqrt(
                9
                * (count_group1_larger + 1)
                * (sample_size - count_group1_larger)
                * (9 * sample_size + 5 - ctag_square)
                + sample_size
                + 1
            )
        )
        D = (
            81 * (count_group1_larger + 1) ** 2
            - 9 * (count_group1_larger + 1) * (2 + ctag_square)
            + 1
        )
        E = 1 + A * ((B + C) / D) ** 3
        A2 = (count_group1_larger / (sample_size - count_group1_larger - 1)) ** 2
        B2 = (
            81 * (count_group1_larger) * (sample_size - count_group1_larger - 1)
            - 9 * sample_size
            - 8
        )
        C2 = (
            3
            * ctag
            * np.sqrt(
                9
                * count_group1_larger
                * (sample_size - count_group1_larger - 1)
                * (9 * sample_size + 5 - ctag_square)
                + sample_size
                + 1
            )
        )
        D2 = (
            81 * count_group1_larger**2
            - (9 * count_group1_larger) * (2 + ctag_square)
            + 1
        )
        E2 = 1 + A2 * ((B2 + C2) / D2) ** 3

        upper_ci_ps_dep_pratt = 1 / E2
        lower_ci_ps_dep_pratt = 1 / E

        if count_group1_larger == 1:
            lower_ci_ps_dep_pratt = 1 - (1 - confidence_level) ** (1 / sample_size)
            upper_ci_ps_dep_pratt = 1 - (confidence_level) ** (1 / sample_size)

        if count_group1_larger == 0:
            lower_ci_ps_dep_pratt = 0
            upper_ci_ps_dep_pratt = beta.ppf(
                1 - confidence_level,
                count_group1_larger + 1,
                sample_size - count_group1_larger,
            )

        if count_group1_larger == sample_size - 1:
            lower_ci_ps_dep_pratt = (confidence_level) ** (1 / sample_size)
            upper_ci_ps_dep_pratt = (1 - confidence_level) ** (1 / sample_size)

        if count_group1_larger == sample_size:
            lower_ci_ps_dep_pratt = (confidence_level * 2) ** (1 / sample_size)
            upper_ci_ps_dep_pratt = 1

        if lower_ci_ps_dep_pratt < 0:
            lower_ci_ps_dep_pratt = 0
        if upper_ci_ps_dep_pratt > 1:
            upper_ci_ps_dep_pratt = 1

        critical_z_value = norm.ppf(0.05 / 2)
        critical_t_value = t.ppf(0.05 / 2, (sample_size - 1))

        feng_standard_error = np.sqrt(
            np.sum((np.sign(difference) - cliffs_delta) ** 2)
            / (sample_size * (sample_size - 1))
        )
        upper_ci_cliff = (
            cliffs_delta
            - cliffs_delta**3
            - critical_t_value
            * feng_standard_error
            * np.sqrt(
                (1 - cliffs_delta**2) ** 2
                + critical_t_value**2 * feng_standard_error**2
            )
        ) / (1 - cliffs_delta**2 + critical_t_value**2 * feng_standard_error**2)
        lower_ci_cliff = (
            cliffs_delta
            - cliffs_delta**3
            + critical_t_value
            * feng_standard_error
            * np.sqrt(
                (1 - cliffs_delta**2) ** 2
                + critical_t_value**2 * feng_standard_error**2
            )
        ) / (1 - cliffs_delta**2 + critical_t_value**2 * feng_standard_error**2)

        number_of_cases_x_larger_than_median_y = sum(
            1 for val in column_1 if val > sample_2_median
        )
        aparametric_cohens_u3_no_ties = (
            number_of_cases_x_larger_than_median_y / sample_size
        )

        if aparametric_cohens_u3_no_ties == 0:
            aparametric_cohens_u3_no_ties = 1 / (sample_size + 1)
        elif aparametric_cohens_u3_no_ties == 1:
            aparametric_cohens_u3_no_ties = sample_size / (sample_size + 1)
        kraemer_andrews_gamma = norm.ppf(aparametric_cohens_u3_no_ties)

        number_of_cases_x_equal_to_median_y = sum(
            1 for val in column_1 if val == sample_2_median
        )
        hentschke_stuttgen_u3 = (
            number_of_cases_x_larger_than_median_y
            + number_of_cases_x_equal_to_median_y * 0.5
        ) / sample_size
        if sample_1_median == sample_2_median:
            hentschke_stuttgen_u3 = 0.5

        number_of_cases_x_larger_than_maximum_y = sum(
            1 for val in column_1 if val > np.max(column_2)
        )
        number_of_cases_x_smaller_than_minimum_y = sum(
            1 for val in column_1 if val < np.min(column_2)
        )
        hentschke_stuttgen_u1 = (
            number_of_cases_x_larger_than_maximum_y
            + number_of_cases_x_smaller_than_minimum_y
        ) / sample_size

        eta = 0

        h1 = max(
            (1.2 * (np.percentile(column_1, 75) - np.percentile(column_1, 25)))
            / (sample_size ** (1 / 5)),
            0.05,
        )
        h2 = max(
            (1.2 * (np.percentile(column_2, 75) - np.percentile(column_2, 25)))
            / (sample_size ** (1 / 5)),
            0.05,
        )

        for value in column_1:
            f_x1 = (
                np.sum(column_1 <= (value + h1)) - np.sum(column_1 < (value - h1))
            ) / (2 * sample_size * h1)
            f_x2 = (
                np.sum(column_2 <= (value + h2)) - np.sum(column_2 < (value - h2))
            ) / (2 * sample_size * h2)
            if f_x1 > f_x2:
                eta += 1

        wilcox_musaka_q_dep = eta / sample_size

        bootstrap_samples_x = []
        for _ in range(reps):
            sample_1_bootstrap = np.random.choice(column_1, len(column_1), replace=True)
            difference_bootstrapping = np.random.choice(
                column_1 - column_2, len(column_1), replace=True
            )
            bootstrap_samples_x.append(sample_1_bootstrap)

        number_of_cases_x_larger_than_median_y_bootstrapping = np.array(
            [(np.sum(sample_x > sample_2_median)) for sample_x in bootstrap_samples_x]
        )
        number_of_cases_x_larger_than_median_y_bootstrapping = (
            number_of_cases_x_larger_than_median_y_bootstrapping / sample_size
        )
        number_of_cases_x_larger_than_median_y_bootstrapping = np.where(
            number_of_cases_x_larger_than_median_y_bootstrapping == 0,
            1 / (sample_size + 1),
            number_of_cases_x_larger_than_median_y_bootstrapping,
        )
        number_of_cases_x_larger_than_median_y_bootstrapping = np.where(
            number_of_cases_x_larger_than_median_y_bootstrapping == 1,
            sample_size / (sample_size + 1),
            number_of_cases_x_larger_than_median_y_bootstrapping,
        )
        kraemer_andrews_gamma_bootstrapping = norm.ppf(
            number_of_cases_x_larger_than_median_y_bootstrapping
        )
        lower_ci_kraemer_andrews_gamma_boot = np.percentile(
            kraemer_andrews_gamma_bootstrapping,
            ((1 - confidence_level) - ((1 - confidence_level) / 2)) * 100,
        )
        upper_ci_kraemer_andrews_gamma_boot = np.percentile(
            kraemer_andrews_gamma_bootstrapping,
            ((confidence_level) + ((1 - confidence_level) / 2)) * 100,
        )

        hentschke_stuttgen_u3_boot = []
        for sample_x in bootstrap_samples_x:
            hentschke_stuttgen_u3_boot.append(
                (
                    np.sum(sample_x > sample_2_median)
                    + np.sum(sample_x == sample_2_median) * 0.5
                )
                / sample_size
            )
            if np.median(sample_x) == sample_2_median:
                hentschke_stuttgen_u3_boot.append(0.5)
        lower_ci_hentschke_stuttgen_u3 = np.percentile(
            hentschke_stuttgen_u3_boot,
            ((1 - confidence_level) - ((1 - confidence_level) / 2)) * 100,
        )
        upper_ci_hentschke_stuttgen_u3 = np.percentile(
            hentschke_stuttgen_u3_boot,
            ((confidence_level) + ((1 - confidence_level) / 2)) * 100,
        )

        number_of_cases_x_larger_than_max_y_bootstrapping = np.array(
            [(np.sum(sample_x > np.max(column_2))) for sample_x in bootstrap_samples_x]
        )
        number_of_cases_x_smaller_than_min_y_bootstrapping = np.array(
            [(np.sum(sample_x < np.min(column_2))) for sample_x in bootstrap_samples_x]
        )
        hentschke_stuttgen_u1_boot = (
            number_of_cases_x_larger_than_max_y_bootstrapping
            + number_of_cases_x_smaller_than_min_y_bootstrapping
        ) / sample_size
        lower_ci_hentschke_stuttgen_u1 = np.percentile(
            hentschke_stuttgen_u1_boot,
            ((1 - confidence_level) - ((1 - confidence_level) / 2)) * 100,
        )
        upper_ci_hentschke_stuttgen_u1 = np.percentile(
            hentschke_stuttgen_u1_boot,
            ((confidence_level) + ((1 - confidence_level) / 2)) * 100,
        )

        cles_dz = res.CLES(
            method="cohen's dz",
            value=float(np.around(cles_dz, 4)),
            standard_error=standard_error_cohens_dz,
            ci_lower=float(np.around(norm.cdf(ci_lower_cohens_dz_central) * 100, 4)),
            ci_upper=float(np.around(norm.cdf(ci_upper_cohens_dz_central) * 100, 4)),
        )
        cles_gz = res.CLES(
            method="hedges' gz",
            value=float(np.around(cles_gz, 4)),
            standard_error=standard_error_hedges_gz,
            ci_lower=float(np.around(norm.cdf(ci_lower_hedges_gz_central) * 100, 4)),
            ci_upper=float(np.around(norm.cdf(ci_upper_hedges_gz_central) * 100, 4)),
        )
        cles_dz.update_pivotal_ci(
            pivotal_ci_lower=float(
                np.around(norm.cdf(ci_lower_cohens_dz_pivotal) * 100, 4)
            ),
            pivotal_ci_upper=float(
                np.around(norm.cdf(ci_upper_cohens_dz_pivotal) * 100, 4)
            ),
        )
        cles_gz.update_pivotal_ci(
            pivotal_ci_lower=float(
                np.around(norm.cdf(ci_lower_hedges_gz_pivotal * correction) * 100, 4)
            ),
            pivotal_ci_upper=float(
                np.around(norm.cdf(ci_upper_hedges_gz_pivotal * correction) * 100, 4)
            ),
        )
        probability_of_superiority = res.ProbabilityOfSuperiority(
            value=round(ps_dep, 4),
            standard_error=0.0,
            ci_lower=np.round(
                min(float(lower_ci_ps_dep_pratt), float(upper_ci_ps_dep_pratt)), 4
            ),
            ci_upper=np.round(
                max(float(lower_ci_ps_dep_pratt), float(upper_ci_ps_dep_pratt)), 4
            ),
        )
        vargha_delaney = res.VarghaDelaney(
            value=round(vda_xy, 4),
            standard_error=0.0,
            ci_lower=round((lower_ci_cliff + 1) / 2, 4),
            ci_upper=round((upper_ci_cliff + 1) / 2, 4),
        )
        cliffs_delta_es = res.CliffsDelta(
            value=round(cliffs_delta, 4),
            standard_error=0.0,
            ci_lower=round(lower_ci_cliff, 4),
            ci_upper=round(upper_ci_cliff, 4),
        )
        kraemer_andrews_gamma = res.KraemerAndrewGamma(
            value=float(kraemer_andrews_gamma),
            standard_error=0.0,
            ci_lower=float(round(lower_ci_kraemer_andrews_gamma_boot, 4)),
            ci_upper=float(round(upper_ci_kraemer_andrews_gamma_boot, 4)),
        )
        non_param_u3 = res.NonParametricU3(
            value=round(hentschke_stuttgen_u3, 4),
            standard_error=0.0,
            ci_lower=float(round(lower_ci_hentschke_stuttgen_u3, 4)),
            ci_upper=float(round(upper_ci_hentschke_stuttgen_u3, 4)),
        )
        non_param_u1 = res.NonParametricU1(
            value=round(hentschke_stuttgen_u1, 4),
            standard_error=0.0,
            ci_lower=float(round(lower_ci_hentschke_stuttgen_u1, 4)),
            ci_upper=float(round(upper_ci_hentschke_stuttgen_u1, 4)),
        )
        wilcox_musaka_q_dep = res.WilcoxMusakaQ(
            value=wilcox_musaka_q_dep,
            standard_error=0.0,
            ci_lower=0.0,
            ci_upper=0.0,
        )

        results = TwoPairedCommonLangResults()
        results.cles_cohen = cles_dz
        results.cles_hedges = cles_gz
        results.probability_of_superiority = probability_of_superiority
        results.vargha_delaney = vargha_delaney
        results.cliff_delta = cliffs_delta_es
        results.kraemer_andrew_gamma = kraemer_andrews_gamma
        results.non_param_u3 = non_param_u3
        results.non_param_u1 = non_param_u1
        results.wilcox_musaka_q = wilcox_musaka_q_dep

        return results

    def pivotal_ci(
        self, t_score, df: float, sample_size: int, confidence_level: float
    ) -> tuple:
        is_negative = False
        if t_score < 0:
            is_negative = True
            t_score = abs(t_score)
        upper_limit = 1 - (1 - confidence_level) / 2
        lower_limit = (1 - confidence_level) / 2

        lower_criterion = [-t_score, t_score / 2, t_score]
        upper_criterion = [t_score, 2 * t_score, 3 * t_score]

        while nct.cdf(t_score, df, lower_criterion[0]) < upper_limit:
            lower_criterion = [
                lower_criterion[0] - t_score,
                lower_criterion[0],
                lower_criterion[2],
            ]

        while nct.cdf(t_score, df, upper_criterion[0]) < lower_limit:
            if nct.cdf(t_score, df) < lower_limit:
                lower_ci = [0, nct.cdf(t_score, df)]
                upper_criterion = [
                    upper_criterion[0] / 4,
                    upper_criterion[0],
                    upper_criterion[2],
                ]

        while nct.cdf(t_score, df, upper_criterion[2]) > lower_limit:
            upper_criterion = [
                upper_criterion[0],
                upper_criterion[2],
                upper_criterion[2] + t_score,
            ]

        lower_ci = 0.0
        diff_lower = 1
        while diff_lower > 0.00001:
            if nct.cdf(t_score, df, lower_criterion[1]) < upper_limit:
                lower_criterion = [
                    lower_criterion[0],
                    (lower_criterion[0] + lower_criterion[1]) / 2,
                    lower_criterion[1],
                ]
            else:
                lower_criterion = [
                    lower_criterion[1],
                    (lower_criterion[1] + lower_criterion[2]) / 2,
                    lower_criterion[2],
                ]
            diff_lower = abs(nct.cdf(t_score, df, lower_criterion[1]) - upper_limit)
            lower_ci = lower_criterion[1] / (np.sqrt(sample_size))

        upper_ci = 0.0
        diff_upper = 1
        while diff_upper > 0.00001:
            if nct.cdf(t_score, df, upper_criterion[1]) < lower_limit:
                upper_criterion = [
                    upper_criterion[0],
                    (upper_criterion[0] + upper_criterion[1]) / 2,
                    upper_criterion[1],
                ]
            else:
                upper_criterion = [
                    upper_criterion[1],
                    (upper_criterion[1] + upper_criterion[2]) / 2,
                    upper_criterion[2],
                ]
            diff_upper = abs(nct.cdf(t_score, df, upper_criterion[1]) - lower_limit)
            upper_ci = upper_criterion[1] / (np.sqrt(sample_size))
        if is_negative:
            return -upper_ci, -lower_ci
        else:
            return lower_ci, upper_ci

    def calculate_central_ci(
        self, cohens_d: float, sample_size: int, confidence_level: float
    ) -> tuple:
        df = sample_size - 1
        correction_factor = math.exp(
            math.lgamma(df / 2)
            - math.log(math.sqrt(df / 2))
            - math.lgamma((df - 1) / 2)
        )
        standard_error_es = np.sqrt(
            (df / (df - 2)) * (1 / sample_size) * (1 + cohens_d**2 * sample_size)
            - (cohens_d**2 / correction_factor**2)
        )
        z_critical_value = norm.ppf(confidence_level + ((1 - confidence_level) / 2))
        ci_lower, ci_upper = (
            cohens_d - standard_error_es * z_critical_value,
            cohens_d + standard_error_es * z_critical_value,
        )
        return ci_lower, ci_upper, standard_error_es



"""
This module contains functions for performing one-sample Z-tests and calculating
various statistics such as Cohen's d, Z-score, p-value, confidence intervals, and
standard errors.

The module includes the following functions:
- calculate_central_ci_from_cohens_d_one_sample: Calculate the confidence intervals
and standard error for Cohen's d effect size in a one-sample Z-test.
- One_Sample_ZTests: A class containing static methods for performing one-sample
Z-tests from Z-score, parameters, and data.
"""

from dataclasses import dataclass
from typing import Optional
import numpy as np
from scipy.stats import norm
from ...utils import interfaces
from ...utils import res


def calculate_central_ci_from_cohens_d_one_sample(
    cohens_d, sample_size, confidence_level
):
    """
    Calculate the confidence intervals and standard error for Cohen's d effect size in a
    one-sample Z-test.

    This function calculates the confidence intervals of the effect size (Cohen's d) for a
    one-sample Z-test or two dependent samples test using the Hedges and Olkin (1985)
    formula to estimate the standard error.

    Parameters
    ----------
    cohens_d : float
        The calculated Cohen's d effect size
    sample_size : int
        The size of the sample
    confidence_level : float
        The confidence level as a decimal (e.g., 0.95 for 95%)

    Returns
    -------
    tuple
        A tuple containing:
        - ci_lower (float): Lower bound of the confidence interval
        - ci_upper (float): Upper bound of the confidence interval
        - standard_error_es (float): Standard error of the effect size

    Notes
    -----
    Since the effect size in the population and its standard deviation are unknown,
    we estimate it based on the sample using the Hedges and Olkin (1985) formula
    to estimate the standard deviation of the effect size.
    """
    standard_error_es = np.sqrt((1 / sample_size) + ((cohens_d**2 / (2 * sample_size))))
    z_critical_value = norm.ppf(confidence_level + ((1 - confidence_level) / 2))
    ci_lower, ci_upper = (
        cohens_d - standard_error_es * z_critical_value,
        cohens_d + standard_error_es * z_critical_value,
    )
    return ci_lower, ci_upper, standard_error_es


@dataclass
class OneSampleZResults:
    """
    A class to store results from one-sample Z statistical tests.

    This class contains attributes to store various statistical measures including:
    - Effect size (Cohen's d)
    - Z-score and p-value
    - Standard error of the mean
    - Confidence intervals for Cohen's d
    - Standard error of the effect size
    """

    cohens_d: Optional[res.CohenD] = None
    z_score: Optional[float] = None
    p_value: Optional[float] = None
    standard_error_of_the_mean: Optional[float] = None
    sample_mean: Optional[float] = None
    population_mean: Optional[float] = None
    population_sd: Optional[float] = None
    sample_sd: Optional[float] = None
    difference_between_means: Optional[float] = None
    sample_size: Optional[float] = None


class OneSampleZTests(interfaces.AbstractTest):
    """
    A class for performing one-sample Z-tests.
    This class provides methods to calculate Z-test results from a Z-score,
    from sample parameters, and from sample data.
    """

    @staticmethod
    def from_score(
        z_score: float, sample_size: float, confidence_level: float = 0.95
    ) -> OneSampleZResults:
        """
        Calculate the one-sample Z-test results from a given Z-score.
        """

        p_value = min(float(norm.sf((abs(z_score))) * 2), 0.99999)
        cohens_d = z_score / np.sqrt(sample_size)
        ci_lower, ci_upper, standard_error_es = (
            calculate_central_ci_from_cohens_d_one_sample(
                cohens_d, sample_size, confidence_level
            )
        )

        cohens_d = res.CohenD(
            value=round(cohens_d, 4),
            ci_lower=round(ci_lower, 4),
            ci_upper=round(ci_upper, 4),
            standard_error=round(standard_error_es, 4),
        )

        results = OneSampleZResults()
        results.z_score = z_score
        results.p_value = p_value
        results.cohens_d = cohens_d

        return results

    @staticmethod
    def from_parameters(
        sample_mean: float,
        sample_size: int,
        population_mean: int,
        population_sd,
        confidence_level=0.95,
    ) -> OneSampleZResults:
        """
        Calculate the one-sample Z-test results from given population and sample parameters.
        """

        mean_standard_error = population_sd / np.sqrt(sample_size)
        z_score = (population_mean - sample_mean) / mean_standard_error
        cohens_d = (population_mean - sample_mean) / population_sd
        p_value = min(float(norm.sf((abs(z_score))) * 2), 0.99999)
        ci_lower, ci_upper, standard_error_es = (
            calculate_central_ci_from_cohens_d_one_sample(
                cohens_d, sample_size, confidence_level
            )
        )

        cohens_d = res.CohenD(
            value=round(cohens_d, 4),
            ci_lower=round(ci_lower, 4),
            ci_upper=round(ci_upper, 4),
            standard_error=round(standard_error_es, 4),
        )

        # Create results object
        results = OneSampleZResults()
        results.z_score = z_score
        results.p_value = p_value
        results.cohens_d = cohens_d
        results.standard_error_of_the_mean = round(mean_standard_error, 4)
        results.sample_mean = sample_mean
        results.population_mean = population_mean
        results.population_sd = population_sd
        return results

    @staticmethod
    def from_data(
        column: list,
        population_mean: float,
        population_sd: float,
        confidence_level: float = 0.95,
    ) -> OneSampleZResults:
        """
        Calculate the one-sample Z-test results from given sample data.
        """

        sample_mean = np.mean(column)
        sample_sd = np.std(column, ddof=1)
        diff_mean = population_mean - sample_mean
        sample_size = len(column)
        standard_error = population_sd / (np.sqrt(sample_size))
        z_score = diff_mean / standard_error
        cohens_d = (
            diff_mean
        ) / population_sd  # This is the effect size for one sample z-test Cohen's d
        p_value = min(float(norm.sf((abs(z_score))) * 2), 0.99999)
        ci_lower, ci_upper, standard_error_es = (
            calculate_central_ci_from_cohens_d_one_sample(
                cohens_d, sample_size, confidence_level
            )
        )

        cohens_d = res.CohenD(
            value=float(cohens_d),
            ci_lower=ci_lower,
            ci_upper=ci_upper,
            standard_error=standard_error_es,
        )

        results = OneSampleZResults()
        results.z_score = z_score
        results.p_value = p_value
        results.cohens_d = cohens_d
        results.sample_size = sample_size
        results.population_mean = population_mean
        results.sample_mean = float(sample_mean)
        results.sample_sd = float(sample_sd)
        results.standard_error_of_the_mean = standard_error
        results.difference_between_means = float(diff_mean)

        return results
